"""
Potentials for orbital density dependent energy functionals
"""

import numpy as np
from gpaw.utilities import pack_density, unpack_hermitian
from gpaw.lfc import LFC
from gpaw.transformers import Transformer
from gpaw.directmin.tools import d_matrix, get_n_occ
from gpaw.poisson import PoissonSolver
import gpaw.cgpaw as cgpaw


class ERLocalization:

    """
    Edmiston-Ruedenberg localisation functions.

    This class return -1/2 sum_i f_i (2integrals rr') n_i n_i'/ (r-r')
    and it is gradients

    """
    def __init__(self, wfs, dens,
                 sic_coarse_grid=True,
                 poisson_solver='FPS'):

        self.name = 'ER_SIC'
        # what we need from wfs
        self.setups = wfs.setups
        spos_ac = wfs.spos_ac
        self.cgd = wfs.gd

        # what we need from dens
        self.finegd = dens.finegd
        self.sic_coarse_grid = sic_coarse_grid

        if self.sic_coarse_grid:
            self.ghat = LFC(self.cgd,
                            [setup.ghat_l for setup
                             in self.setups],
                            integral=np.sqrt(4 * np.pi),
                            forces=True)
            self.ghat.set_positions(spos_ac)
        else:
            self.ghat = dens.ghat  # we usually solve poiss. on finegd

        if poisson_solver == 'FPS':
            self.poiss = PoissonSolver(use_charge_center=True,
                                       use_charged_periodic_corrections=True)
        elif poisson_solver == 'GS':
            self.poiss = PoissonSolver(name='fd',
                                       relax=poisson_solver,
                                       eps=1.0e-16,
                                       use_charge_center=True,
                                       use_charged_periodic_corrections=True)

        if self.sic_coarse_grid is True:
            self.poiss.set_grid_descriptor(self.cgd)
        else:
            self.poiss.set_grid_descriptor(self.finegd)

        self.interpolator = Transformer(self.cgd, self.finegd, 3)
        self.restrictor = Transformer(self.finegd, self.cgd, 3)
        self.dtype = wfs.dtype
        self.eigv_s = {}
        self.lagr_diag_s = {}
        self.e_sic_by_orbitals = {}
        self.counter = 0  # number of calls of this class

        self.n_kps = wfs.kd.nibzkpts

    def get_orbdens_compcharge_dm_kpt(self, wfs, kpt, n):

        if wfs.mode == 'pw':
            nt_G = wfs.pd.gd.zeros(global_array=True)
            psit_G = wfs.pd.alltoall1(kpt.psit.array[n:n + 1], kpt.q)
            if psit_G is not None:
                psit_R = wfs.pd.ifft(psit_G, kpt.q,
                                     local=True, safe=False)
                cgpaw.add_to_density(1.0, psit_R, nt_G)
            wfs.pd.gd.comm.sum(nt_G)
            nt_G = wfs.pd.gd.distribute(nt_G)
        else:
            nt_G = np.absolute(kpt.psit_nG[n] ** 2)

        # paw
        Q_aL = {}
        D_ap = {}
        for a, P_ni in kpt.P_ani.items():
            P_i = P_ni[n]
            D_ii = np.outer(P_i, P_i.conj()).real
            D_ap[a] = D_p = pack_density(D_ii)
            Q_aL[a] = np.dot(D_p, self.setups[a].Delta_pL)

        return nt_G, Q_aL, D_ap

    def get_energy_and_gradients_kpt(self, wfs, kpt, grad_knG):

        wfs.timer.start('SIC e/g grid calculations')
        k = self.n_kps * kpt.s + kpt.q
        n_occ = get_n_occ(kpt)[0]

        e_total_sic = np.array([])

        for i in range(n_occ):
            nt_G, Q_aL, D_ap = \
                self.get_orbdens_compcharge_dm_kpt(wfs, kpt, i)

            # calculate - SI Hartree
            e_sic, v_ht_g = \
                self.get_pseudo_pot(nt_G, Q_aL, i, kpoint=k)

            # calculate PAW corrections
            e_sic_paw_m, dH_ap = \
                self.get_paw_corrections(D_ap, v_ht_g)

            if wfs.mode == 'pw':
                v_ht_g = wfs.pd.gd.collect(v_ht_g, broadcast=True)
                Q_G = wfs.pd.Q_qG[kpt.q]
                psit_G = wfs.pd.alltoall1(kpt.psit_nG[i: i + 1], kpt.q)
                if psit_G is not None:
                    psit_R = wfs.pd.ifft(
                        psit_G, kpt.q, local=True, safe=False)
                    psit_R *= v_ht_g
                    wfs.pd.fftplan.execute()
                    vtpsit_G = wfs.pd.tmp_Q.ravel()[Q_G]
                else:
                    vtpsit_G = wfs.pd.tmp_G
                wfs.pd.alltoall2(vtpsit_G, kpt.q, grad_knG[k][i: i + 1])
                grad_knG[k][i] *= kpt.f_n[i]
            else:
                grad_knG[k][i] += kpt.psit_nG[i] * v_ht_g * kpt.f_n[i]

            # total sic:
            e_sic += e_sic_paw_m
            c_axi = {}
            for a in kpt.P_ani.keys():
                dH_ii = unpack_hermitian(dH_ap[a])
                c_xi = np.dot(kpt.P_ani[a][i], dH_ii)
                c_axi[a] = c_xi * kpt.f_n[i]
            # add projectors to
            wfs.pt.add(grad_knG[k][i], c_axi, kpt.q)

            e_total_sic = np.append(e_total_sic,
                                    kpt.f_n[i] * e_sic, axis=0)

        self.e_sic_by_orbitals[k] = np.copy(e_total_sic)
        wfs.timer.stop('SIC e/g grid calculations')

        return e_total_sic.sum()

    def get_pseudo_pot(self, nt, Q_aL, i, kpoint=None):

        if self.sic_coarse_grid is False:
            v_ht_g = self.finegd.zeros()
            nt_sg = self.finegd.zeros()
        else:
            v_ht_g = self.cgd.zeros()
            nt_sg = self.cgd.zeros()

        if self.sic_coarse_grid is False:
            self.interpolator.apply(nt, nt_sg)
            nt_sg *= self.cgd.integrate(nt) / \
                self.finegd.integrate(nt_sg)
        else:
            nt_sg = nt

        self.ghat.add(nt_sg, Q_aL)

        self.poiss.solve(v_ht_g, nt_sg, zero_initial_phi=False)

        if self.sic_coarse_grid is False:
            ec = 0.5 * self.finegd.integrate(nt_sg * v_ht_g)
        else:
            ec = 0.5 * self.cgd.integrate(nt_sg * v_ht_g)

        return np.array([-ec]), -1.0 * v_ht_g

    def get_paw_corrections(self, D_ap, vHt_g):

        # Hartree-PAW
        dH_ap = {}
        for a, D_p in D_ap.items():
            dH_ap[a] = np.zeros(len(D_p))

        ec = 0.0
        W_aL = self.ghat.dict()
        self.ghat.integrate(vHt_g, W_aL)
        for a, D_p in D_ap.items():
            setup = self.setups[a]
            M_p = np.dot(setup.M_pp, D_p)
            ec += np.dot(D_p, M_p)
            dH_ap[a] += (2.0 * M_p + np.dot(setup.Delta_pL, W_aL[a]))
        if self.sic_coarse_grid is False:
            ec = self.finegd.comm.sum_scalar(ec)
        else:
            ec = self.cgd.comm.sum_scalar(ec)

        return np.array([-ec]), dH_ap

    def get_energy_and_gradients_inner_loop(
            self, wfs, kpt, a_mat, evals, evec):

        e_sic, l_odd = \
            self.get_energy_and_hamiltonian_kpt(wfs, kpt)
        wfs.timer.start('Unitary gradients')
        f = np.ones(l_odd.shape[0])

        # calc_error:
        indz = np.absolute(l_odd) > 1.0e-4
        l_c = 2.0 * l_odd[indz]
        l_odd = f[:, np.newaxis] * l_odd.T.conj() - f * l_odd
        kappa = np.max(np.absolute(l_odd[indz]) / np.absolute(l_c))

        if a_mat is None:
            wfs.timer.stop('Unitary gradients')
            return l_odd.T.conj(), e_sic, kappa
        else:
            g_mat = evec.T.conj() @ l_odd.T.conj() @ evec
            g_mat = g_mat * d_matrix(evals)
            g_mat = evec @ g_mat @ evec.T.conj()

            for i in range(g_mat.shape[0]):
                g_mat[i][i] *= 0.5

            wfs.timer.stop('Unitary gradients')
            if a_mat.dtype == float:
                g_mat = g_mat.real
            return 2.0 * g_mat, e_sic, kappa

    def get_energy_and_hamiltonian_kpt(self, wfs, kpt):

        n_occ = get_n_occ(kpt)[0]
        k = self.n_kps * kpt.s + kpt.q
        grad = {k: np.zeros_like(kpt.psit_nG[:n_occ])}

        e_sic = self.get_energy_and_gradients_kpt(wfs, kpt, grad)

        l_odd = wfs.integrate(kpt.psit_nG[:n_occ],
                              grad[k][:n_occ], True)

        return e_sic, l_odd



from __future__ import annotations
import numpy as np
from gpaw.typing import Array1D


def sampling_branches(w_dist: Array1D,
                      parallel_lines: int = 2,
                      varpi: float = 1,
                      eta0: float = 1e-5,
                      eta_rest: float = 0.1) -> Array1D:
    """
        w_dist         an array of points in real axis
        parallel_lines How many lines to parallel to the real frequency axis
                       the sampling has.
        varpi          The distance of the second line from the real axis
        d              [d0, drest], where d0 is the imaginary part of the
                       first point of the first line, and drest is the
                       imaginary part of the rest of the points of the first
                       line.
    """
    if parallel_lines not in [1, 2]:
        raise ValueError('parallel_lines must be either 1 or 2.')

    if len(w_dist) == 1:
        assert eta0 >= 0
        assert parallel_lines == 2
        w_grid = np.array([w_dist + 1j * eta0, w_dist + 1j * varpi],
                          dtype=complex)
        return w_grid

    if parallel_lines == 1:  # only one branch
        assert varpi >= 0
        w_grid = w_dist + 1j * varpi
        return w_grid

    # parallel lines == 2
    assert eta0 >= 0
    assert eta_rest >= 0
    assert varpi > eta0 and varpi > eta_rest
    w_grid = np.concatenate((np.array([w_dist[0] + 1j * eta0]),
                            w_dist[1:] + 1j * eta_rest, w_dist + 1j * varpi))
    return w_grid


def frequency_distribution(npoles: int,
                           wrange: tuple[float, float] | Array1D,
                           alpha: float = 1) -> Array1D:

    if npoles == 1:
        w_grid = np.array([0.])
        return w_grid

    assert wrange[0].real >= 0
    if not wrange[1].real > wrange[0].real:
        raise ValueError('Frequency range inverted.')

    if alpha == 0:  # homogeneous distribution
        w_grid = np.linspace(wrange[0], wrange[1], 2 * npoles)
        return w_grid

    ws = wrange[1] - wrange[0]
    w_grid = wrange[0] + ws * semi_homogenous_partition(npoles)**alpha
    return w_grid


def mpa_frequency_sampling(npoles: int,
                           wrange: tuple[float, float],
                           varpi: float,
                           eta0: float,
                           eta_rest: float,
                           parallel_lines: int = 2,
                           alpha: float = 1) -> Array1D:
    """
    This function creates a frequency grid in the complex plane.
    The grid can have 1 or 2 branches with points non homogeneously
    distributed along the real frequency axis.
    See Fig. 1 and Eq. (18) of Ref. [1], and Eq. (11) of Ref. [2]
    [1] DA. Leon et al, PRB 104, 115157 (2021)
    [2] DA. Leon et al, PRB 107, 155130 (2023)

    Parameters
    ----------
    npoles : numper of poles (half the number of frequency points)
    wrange : [w_start, w_end]. An array of two complex numbers defining
             the sampling range
    varpi : immaginary part of the second branch
    eta0 : damping factor for the first point of the first branch
    eta_rest : damping for the rest of the points of the first branch
    parallel_lines : Either 1 or 2, how many lines there are parallel
                     to the real axis.
    alpha : exponent of the distribution of points along the real axis
    ______________________________________________________________________
                  Example: double parallel sampling with 9 poles
    ----------------------------------------------------------------------
                            complex frequency plane
    ----------------------------------------------------------------------
    |     (wrange[0], varpi) ... . . . . .   . (wrange[1], varpi)        |
    |                                                                    |
    |                         .. . . . . .   . (wrange[1], eta_rest)     |
    |      (wrange[0], eta0) .                                           |
    |____________________________________________________________________|
    """
    _wrange = np.array(wrange)
    grid_p = frequency_distribution(npoles, _wrange, alpha)
    grid_w = sampling_branches(grid_p, parallel_lines=parallel_lines,
                               varpi=varpi, eta0=eta0, eta_rest=eta_rest)
    return grid_w


def semi_homogenous_partition(npoles: int) -> Array1D:
    """
    Returns a semi-homogenous partition with n-poles between 0 and 1
    according to Eq. (18) of Ref. [1], and Eq. (11) of Ref. [2]
    [1] DA. Leon et al, PRB 104, 115157 (2021)
    [2] DA. Leon et al, PRB 107, 155130 (2023)
    """
    small_cases = {1: np.array([0.0]),
                   2: np.array([0.0, 1.0]),
                   3: np.array([0.0, 0.5, 1.0])}
    if npoles < 4:
        return small_cases[npoles]
    # Calculate the grid spacing
    # Round up to the next power of 2. This will determine the minimum spacing
    dw = 1 / 2**np.ceil(np.log2(npoles))
    dw_n = np.zeros(npoles)

    # Get the previous power of two, by searching down,
    # e.g. lp(4) = 2, lp(7)=4, lp(8)=4, lp(9)=8
    lp = 2**int(np.log2(npoles - 1))

    # There are usually 2 kinds of intervals in a semi homogenous grid,
    # they are always in order such that smallest intervals are closer to zero.
    # The interval sizes are dw, 2*dw.
    # There is an exception to this rule when npoles == power of two + 1,
    # because then there would be only one type of interval with the rule
    # below. To keep the grid inhomogenous, one adds the third interval 4 * dw.
    if npoles == lp + 1:
        np1 = 2
        np3 = 1
    else:
        np1 = (npoles - (lp + 1)) * 2
        np3 = 0
    # The number of intervals is always one less, than the number of points
    # in the grid. Therefore, We can deduce np2 from np1 and np3.
    np2 = npoles - 1 - np1 - np3

    # Build up the intervals onto an array
    dw_n = np.repeat([0, 1, 2, 4], [1, np1, np2, np3])

    # Sum over the intervals to build the point grid
    w_grid = np.cumsum(dw_n) * dw
    return w_grid



import pytest

from ase import Atoms
from ase.build import fcc111, graphene_nanoribbon

from gpaw.utilities.adjust_cell import adjust_cell
from gpaw.utilities import h2gpts
from gpaw.grid_descriptor import GridDescriptor


def test_non_periodic():
    R = 2.0
    b = 4.0
    h = 0.2

    CO = Atoms(['C', 'O'], [(1, 0, 0), (1, 0, R)])

    adjust_cell(CO, b, h)
    cc = CO.get_cell()

    for c in range(3):
        width = 2 * b
        if c == 2:
            width += R + 2 * h
        assert cc[c, c] == pytest.approx(width, abs=1e-10)


def test_non_orthogonal_unitcell():
    a = 3.912
    box = 3.
    h = 0.2

    for atoms in [
            fcc111('Pt', (1, 1, 1), a=a),
            fcc111('Pt', (5, 6, 2), a=a, orthogonal=True)]:
        old_cell = atoms.cell.copy()

        adjust_cell(atoms, box, h)

        # check that the box ajusts for h in only non periodic directions
        assert atoms.cell[:2, :2] == pytest.approx(old_cell[:2, :2])
        # check that the atom is shifted in non periodic direction
        assert (atoms.positions[:, 2] >= box).all()

        N_c = h2gpts(h, atoms.cell)
        gd = GridDescriptor(N_c, atoms.cell, atoms.pbc)
        h_c = gd.get_grid_spacings()

        assert h_c[2] == pytest.approx(h_c[:2].sum() / 2)


def test_rotated_unitcell():
    h = 0.2
    box = 3

    gnt = graphene_nanoribbon(2, 2, sheet=True, vacuum=3)

    gnt.rotate('y', 'z', rotate_cell=True)

    adjust_cell(gnt, box, h)

    # check if the atoms are inside the unitcell
    gnt2 = gnt.copy()
    gnt2.pbc = False
    spos = gnt2.get_scaled_positions()

    assert ((spos) <= 1).all() and (spos >= 0).all()
    # check that the vacuum is equal on both sides in non periodic direction
    cell_z = gnt.cell[1, 2]
    assert (gnt.positions[:, 2] == cell_z / 2).all()

    N_c = h2gpts(h, gnt.cell)
    gd = GridDescriptor(N_c, gnt.cell, gnt.pbc)
    h_c = gd.get_grid_spacings()

    assert h_c[1] == pytest.approx((h_c[0] + h_c[2]) / 2)



"""Check for tunability of gamma for yukawa potential."""
import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.utilities.adjust_cell import adjust_cell
from gpaw.eigensolvers import RMMDIIS
from gpaw.xc.hybrid import HybridXC
from gpaw.occupations import FermiDirac
from gpaw.test import gen
import gpaw.cgpaw as cgpaw


@pytest.mark.hybrids
def test_rsf_yukawa_rsf_general(in_tmp_dir, add_cwd_to_setup_paths):
    libxc_version = getattr(cgpaw, 'libxc_version', '2.x.y')
    if int(libxc_version.split('.')[0]) < 3:
        from unittest import SkipTest
        raise SkipTest

    for atom in ['Be']:
        gen(atom, xcname='PBE', scalarrel=True, exx=True,
            yukawa_gamma=0.83, gpernode=149)

    h = 0.35
    be = Atoms('Be', positions=[[0, 0, 0]])
    adjust_cell(be, 3.0, h=h)

    c = {'energy': 0.05, 'eigenstates': 0.05, 'density': 0.05}

    IP = 8.79

    xc = HybridXC('LCY-PBE', omega=0.83)
    fname = 'Be_rsf.gpw'

    calc = GPAW(mode='fd', txt='Be.txt', xc=xc, convergence=c,
                eigensolver=RMMDIIS(), h=h,
                occupations=FermiDirac(width=0.0), spinpol=False)
    be.calc = calc
    energy_083 = be.get_potential_energy()
    (eps_homo, eps_lumo) = calc.get_homo_lumo()
    assert eps_homo == pytest.approx(-IP, abs=0.15)
    xc2 = 'LCY-PBE'
    energy_075 = calc.get_xc_difference(HybridXC(xc2))
    assert energy_083 - energy_075 == pytest.approx(21.13, abs=0.2)
    calc.write(fname)
    calc2 = GPAW(fname)
    xc = calc2.hamiltonian.xc
    assert xc.name == 'LCY-PBE', 'wrong name for functional'
    assert xc.omega == 0.83, 'wrong value for RSF omega'



from __future__ import annotations

import numpy as np
from ase.units import Ha, Bohr

from gpaw.typing import Array1D, Array2D, ArrayLike1D
from gpaw.core import UGArray
from gpaw.new.density import Density


def create_external_potential(params: dict) -> ExternalPotential:
    if not params:
        return ExternalPotential()
    params = params.copy()
    name = params.pop('name')
    if name == 'BField':
        return BField(**params)
    if name == 'ConstantElectricField':
        return ConstantElectricField(**params)
    raise ValueError


class ExternalPotential:
    def update_potential(self,
                         vt_sR: UGArray,
                         density) -> float:
        return 0.0

    def add_paw_correction(self, Delta_p: Array1D, dH_sp: Array2D) -> float:
        return 0.0


class ConstantElectricField(ExternalPotential):
    def __init__(self, strength, direction=[0, 0, 1], tolerance=1e-7):
        """External constant electric field.

        strength: float
            Field strength in V/Ang.
        direction: vector
            Polarization direction.
        """
        self.strength = strength * Bohr / Ha
        self.direction_v = np.array(direction, dtype=float)
        self.direction_v /= np.linalg.norm(self.direction_v)
        self.field_v = self.strength * self.direction_v
        self.tolerance = tolerance
        self.name = 'ConstantElectricField'

    def __str__(self):
        return ('Constant electric field: '
                '({:.3f}, {:.3f}, {:.3f}) V/Ang'
                .format(*(self.field_v * Ha / Bohr)))

    def update_potential(self,
                         vt_sR: UGArray,
                         density) -> float:
        grid = vt_sR.desc
        L_c = grid.cell_cv @ self.direction_v
        (axis,) = np.where(abs(L_c) > self.tolerance)[0]
        # assert not grid.pbc_c[axis]
        fracpos_r = np.linspace(grid.start_c[axis],
                                grid.end_c[axis],
                                grid.mysize_c[axis],
                                endpoint=False) / grid.size_c[axis]
        v_r = L_c[axis] * (fracpos_r - 0.5) * self.strength
        if grid.start_c[axis] == 0:
            v_r[0] = 0.0
        vt_sR.data += v_r.reshape([1] +
                                  [-1 if c == axis else 1 for c in range(3)])
        return 0.0


class BField(ExternalPotential):
    def __init__(self, field: ArrayLike1D):
        """Constant magnetic field.

        field:
            B-field vector in units of Ha/bohr-magnoton.
        """
        self.field_v = np.array(field) / Ha
        assert self.field_v.shape == (3,)

    def update_potential(self,
                         vt_sR: UGArray,
                         density: Density) -> float:
        magmom_v, _ = density.calculate_magnetic_moments()
        eext = -self.field_v @ magmom_v
        ncomponents = len(vt_sR)
        if ncomponents == 2:
            assert (self.field_v[:2] == 0.0).all()
            vt_sR.data[0] -= self.field_v[2]
            vt_sR.data[1] += self.field_v[2]
        elif ncomponents == 4:
            vt_sR.data[1:] = -self.field_v.reshape((3, 1, 1, 1))
        else:
            1 / 0
        return eext

    def add_paw_correction(self, Delta_p: Array1D, dH_sp: Array2D) -> float:
        if len(dH_sp) == 2:
            c = (4 * np.pi)**0.5 * self.field_v[2]
            dH_sp[0] -= c * Delta_p
            dH_sp[1] += c * Delta_p
        else:
            c_vp = (4 * np.pi)**0.5 * self.field_v[:, np.newaxis]
            dH_sp[1:] -= c_vp * Delta_p
        return 0.0



import pytest
import numpy as np
from gpaw.mpi import world
from gpaw.test import findpeak
from gpaw.response.bse import BSE
from gpaw.response.df import read_response_function


def create_bse(gpwfile, q_c=(0, 0, 0)):
    bse = BSE(gpwfile,
              q_c=q_c,
              spinors=True,
              ecut=10,
              valence_bands=[8],
              conduction_bands=[9],
              eshift=0.8,
              nbands=15,
              mode='BSE',
              truncation='2D')
    return bse


@pytest.mark.response
def test_response_bse_MoS2_cut(in_tmp_dir, scalapack, gpw_files):
    gpwfile = gpw_files['mos2_5x5_pw']
    bse = create_bse(gpwfile)

    outw_w, outalpha_w = bse.get_polarizability(write_eig=None,
                                                eta=0.02,
                                                w_w=np.linspace(0., 5., 5001))
    world.barrier()
    w_w, alphareal_w, alphaimag_w = read_response_function('pol_bse.csv')

    # Check consistency with written results
    assert np.allclose(outw_w, w_w, atol=1e-5, rtol=1e-4)
    assert np.allclose(outalpha_w.real, alphareal_w, atol=1e-5, rtol=1e-4)
    assert np.allclose(outalpha_w.imag, alphaimag_w, atol=1e-5, rtol=1e-4)

    w0, I0 = findpeak(w_w[:1100], alphaimag_w[:1100])
    w1, I1 = findpeak(w_w[1100:1300], alphaimag_w[1100:1300])
    w1 += 1.1

    assert w0 == pytest.approx(0.58, abs=0.01)
    assert I0 == pytest.approx(38.8, abs=0.35)
    assert w1 == pytest.approx(2.22, abs=0.01)
    assert I1 == pytest.approx(6.3, abs=0.35)

    #################################################################
    # Absorption and EELS spectra for 2D materials should be identical
    # for q=0.
    #################################################################

    bse = create_bse(gpwfile)
    outw_w, eels = bse.get_eels_spectrum(w_w=w_w)

    bse = create_bse(gpwfile)
    factor = bse.gs.cd.nonperiodic_hypervolume / (4 * np.pi)
    outw_w, pol = bse.get_polarizability(w_w=w_w)
    assert np.allclose(pol.imag / factor, eels)

    #####################################################################
    # Absorption and EELS spectra for 2D materials should NOT be identical
    # for finite q.
    #####################################################################
    bse = create_bse(gpwfile, q_c=[0.2, 0.2, 0.0])
    outw_w, eels = bse.get_eels_spectrum(w_w=w_w)
    bse = create_bse(gpwfile, q_c=[0.2, 0.2, 0.0])
    outw_w, pol = bse.get_polarizability(w_w)

    assert not np.allclose(pol.imag / factor, eels)



import numpy as np
import pytest
from ase.build import molecule

from gpaw import GPAW
from gpaw.lcao.projected_wannier import get_lcao_projections_HSP
from gpaw.poisson import FDPoissonSolver


@pytest.mark.later
def test_lcao_lcao_projections():
    atoms = molecule('C2H2')
    atoms.center(vacuum=3.0)
    calc = GPAW(mode='fd',
                gpts=(32, 32, 48),
                poissonsolver=FDPoissonSolver(),
                eigensolver='rmm-diis')
    atoms.calc = calc
    atoms.get_potential_energy()

    V_qnM, H_qMM, S_qMM, P_aqMi = get_lcao_projections_HSP(
        calc, bfs=None, spin=0, projectionsonly=False)

    # Test H and S
    eig = sorted(np.linalg.eigvals(np.linalg.solve(S_qMM[0], H_qMM[0])).real)
    eig_ref = np.array([-17.87913099255579, -13.24870583835115,
                        -11.431118704888123, -7.125564231198733,
                        -7.1255642311987195, 0.5929813570452659,
                        0.5929813570454503, 3.925217670277378,
                        7.451140780537926, 26.734705668744386])
    print(eig)
    assert np.allclose(eig, eig_ref)



"""
==  ==========
R
r
G
g
h
x   r or h
==  ==========

"""

from __future__ import annotations

from collections import defaultdict
from typing import DefaultDict

import numpy as np
from gpaw.core.arrays import DistributedArrays
from gpaw.core.atom_arrays import AtomArrays
from gpaw.core.uniform_grid import UGArray
from gpaw.new import trace, zips
from gpaw.new.potential import Potential
from gpaw.new.xc import Functional
from gpaw.setup import Setup
from gpaw.spinorbit import soc as soc_terms
from gpaw.typing import Array1D, Array2D, Array3D
from gpaw.utilities import pack_hermitian, pack_density, unpack_hermitian
from gpaw.yml import indent
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new.external_potential import ExternalPotential


class PotentialCalculator:
    def __init__(self,
                 xc: Functional,
                 poisson_solver,
                 setups: list[Setup],
                 *,
                 fracpos_ac: Array2D,
                 external_potential: ExternalPotential | None = None,
                 soc: bool = False):
        self.poisson_solver = poisson_solver
        self.xc = xc
        self.setups = setups
        self.external_potential = external_potential or ExternalPotential()
        self.fracpos_ac = fracpos_ac
        self.soc = soc

    def __str__(self):
        return (f'{self.poisson_solver}\n'
                f'xc functional:\n{indent(self.xc)}\n')

    def calculate_pseudo_potential(self,
                                   density,
                                   ibzwfs,
                                   vHt_x: DistributedArrays | None
                                   ) -> tuple[dict[str, float],
                                              UGArray,
                                              UGArray,
                                              DistributedArrays]:
        raise NotImplementedError

    def calculate_charges(self, vHt_x):
        raise NotImplementedError

    def restrict(self, a_r, a_R=None):
        raise NotImplementedError

    def calculate_without_orbitals(self,
                                   density,
                                   ibzwfs=None,
                                   vHt_x: DistributedArrays | None = None,
                                   kpt_band_comm: MPIComm | None = None
                                   ) -> tuple[Potential, AtomArrays]:
        xc = self.xc
        if xc.exx_fraction != 0.0:
            from gpaw.new.xc import create_functional
            self.xc = create_functional('PBE', xc.grid)
        potential, Q_al = self.calculate(density, ibzwfs, vHt_x, kpt_band_comm)
        if xc.exx_fraction != 0.0:
            self.xc = xc
        return potential, Q_al

    def calculate(self,
                  density,
                  ibzwfs=None,
                  vHt_x: DistributedArrays | None = None,
                  kpt_band_comm: MPIComm | None = None
                  ) -> tuple[Potential, AtomArrays]:
        energies, vt_sR, dedtaut_sr, vHt_x = self.calculate_pseudo_potential(
            density, ibzwfs, vHt_x)

        e_kinetic = 0.0
        for spin, (vt_R, nt_R) in enumerate(zips(vt_sR, density.nt_sR)):
            e_kinetic -= vt_R.integrate(nt_R)
            if spin < density.ndensities:
                e_kinetic += vt_R.integrate(density.nct_R)

        if dedtaut_sr is not None:
            dedtaut_sR = self.restrict(dedtaut_sr)
            for dedtaut_R, taut_R in zips(dedtaut_sR,
                                          density.taut_sR):
                e_kinetic -= dedtaut_R.integrate(taut_R)
                e_kinetic += dedtaut_R.integrate(density.tauct_R)
        else:
            dedtaut_sR = None

        energies['kinetic'] = e_kinetic

        if kpt_band_comm is None:
            if ibzwfs is None:
                kpt_band_comm = serial_comm
            else:
                kpt_band_comm = ibzwfs.kpt_band_comm
        Q_aL = self.calculate_charges(vHt_x)
        dH_asii, corrections = calculate_non_local_potential(
            self.setups,
            density,
            self.xc,
            self.external_potential,
            Q_aL,
            self.soc,
            kpt_band_comm)

        energies['spinorbit'] = 0.0
        for key, e in corrections.items():
            if 0:
                print(f'{key:10} {energies[key]:15.9f} {e:15.9f}')
            energies[key] += e

        return Potential(vt_sR, dH_asii, dedtaut_sR, energies, vHt_x), Q_aL


@trace
def calculate_non_local_potential(setups,
                                  density,
                                  xc,
                                  ext_pot,
                                  Q_aL,
                                  soc: bool,
                                  kpt_band_comm: MPIComm
                                  ) -> tuple[AtomArrays,
                                             dict[str, float]]:
    dtype = float if density.ncomponents < 4 else complex
    D_asii = density.D_asii.to_xp(np)
    dH_asii = D_asii.layout.new(dtype=dtype).empty(density.ncomponents)
    Q_aL = Q_aL.to_xp(np)
    energy_corrections: DefaultDict[str, float] = defaultdict(float)
    rank = 0
    for a, D_sii in D_asii.items():
        if rank % kpt_band_comm.size == kpt_band_comm.rank:
            Q_L = Q_aL[a]
            setup = setups[a]
            dH_sii, corrections = calculate_non_local_potential1(
                setup, xc, ext_pot, D_sii, Q_L, soc)
            dH_asii[a][:] = dH_sii
            for key, e in corrections.items():
                energy_corrections[key] += e
        else:
            dH_asii[a][:] = 0.0
        rank += 1

    kpt_band_comm.sum(dH_asii.data)

    # Sum over domain:
    names = ['kinetic', 'coulomb', 'zero', 'xc', 'external', 'spinorbit']
    energies = np.array([energy_corrections[name] for name in names])
    density.D_asii.layout.atomdist.comm.sum(energies)
    kpt_band_comm.sum(energies)

    return (dH_asii.to_xp(density.D_asii.layout.xp),
            dict(zips(names, energies)))


def calculate_non_local_potential1(setup: Setup,
                                   xc: Functional,
                                   ext_pot,
                                   D_sii: Array3D,
                                   Q_L: Array1D,
                                   soc: bool) -> tuple[Array3D,
                                                       dict[str, float]]:
    ncomponents = len(D_sii)
    ndensities = 2 if ncomponents == 2 else 1
    D_sp = np.array([pack_density(D_ii.real) for D_ii in D_sii])

    D_p = D_sp[:ndensities].sum(0)

    dH_p = (setup.K_p + setup.M_p +
            setup.MB_p + 2.0 * setup.M_pp @ D_p +
            setup.Delta_pL @ Q_L)
    e_kinetic = setup.K_p @ D_p + setup.Kc
    e_zero = setup.MB + setup.MB_p @ D_p
    e_coulomb = setup.M + D_p @ (setup.M_p + setup.M_pp @ D_p)

    dH_sp = np.zeros_like(D_sp, dtype=float if ncomponents < 4 else complex)

    e_soc = 0.
    if soc:
        dHsoc_sii = soc_terms(setup, xc.xc, D_sp)
        e_soc += (D_sii[1:4] * dHsoc_sii).sum().real
        dH_sp[1:4] = pack_hermitian(dHsoc_sii)

    dH_sp[:ndensities] = dH_p
    e_xc = xc.calculate_paw_correction(setup, D_sp, dH_sp)

    e_external = ext_pot.add_paw_correction(setup.Delta_pL[:, 0], dH_sp)

    dH_sii = unpack_hermitian(dH_sp)

    if setup.hubbard_u is not None:
        eU, dHU_sii = setup.hubbard_u.calculate(setup, D_sii)
        e_xc += eU
        dH_sii += dHU_sii

    e_kinetic -= (D_sii * dH_sii).sum().real

    return dH_sii, {'kinetic': e_kinetic,
                    'coulomb': e_coulomb,
                    'zero': e_zero,
                    'xc': e_xc,
                    'external': e_external,
                    'spinorbit': e_soc}



import numpy as np

import pytest
from gpaw.lrtddft.apmb import sqrt_matrix

# check sqrt of a matrix


def test_linalg_lapack():
    A = [[20, 4], [4, 1]]
    a = [[4.4, 0.8], [0.8, 0.6]]
    A = np.array(A, float)
    print('A=', A)
    a = np.array(a)
    b = sqrt_matrix(A)
    print('sqrt(A)=', b)
    assert ((a - b)**2).sum() == pytest.approx(0, abs=1.0e-12)



r"""
    Objective function class for
    Generalized Pipek-Mezey orbital localization.

    Given a spin channel index the objective function is:
           __ __
           \  \  | A  |p
    P(W) = /  /  |Q(W)|           Eq.1
           -- -- | ii |
            A  i

    where p is a penalty degree: p>1, p<1, not p=1,
    (note that p<1 corresponds to minimization)
    and
           __
     A     \     A
    Q(W) = / W* Q  W              Eq.2
     jj    -- rj rs sj
           rs

    rs run over occupied states only.

     A
    Q    can be defined with two methods:
     rs

    Hirshfeld scheme: 'H'

     A    /    *   A
    Q   = | Phi(r)w(r)Phi(r) dr   Eq.4
     rs   /    r         s

          A
    with w(r) a weight function with center on atom A.
     A
    w(r) is constructed from simple and general gaussians.

    and Wigner-Seitz scheme: 'W'

     A    /    *   A
    Q   = | Phi(r)O(r)Phi(r) dr   Eq.5
     rs   /    r         s

          A              A      B
    with O(r) = 1 if |r-R |>|r-R |, 0 otherwise

    All integrals are performed over the course gd.

"""
import numpy as np
from scipy.linalg import inv, sqrtm
from math import pi
from ase.transport.tools import dagger
from gpaw.pipekmezey.weightfunction import WeightFunc, WignerSeitz
from gpaw.pipekmezey.wannier_basic import md_min, get_atoms_object_from_wfs
from ase.dft.wannier import calculate_weights
from ase.dft.kpoints import get_monkhorst_pack_size_and_offset
from ase.parallel import world


def random_orthogonal(rng, s, dtype=float):
    # Make a random orthogonal matrix of dim s x s,
    # such that WW* = I = W*W
    w_r = rng.random((s, s))
    if dtype == complex:
        w_r = w_r + 1.j * rng.random((s, s))
    return w_r.dot(inv(sqrtm(w_r.T.conj().dot(w_r))))


class PipekMezey:
    """ General Pipek-Mezey Wannier functions:
        J. Chem. Theory Comput. 2017, 13, 2, 460–474

        Parameters
        ----------
        wfs  : GPAW wfs object
        calc : GPAW calculator object

        method : string
           'W' Wigner-Seitz or 'H' Hirshfeld

        penalty : int
           positive (int) value for maximization (localization)
           negative (int) value for minimization (delocalized)

        spin  : int
           spin channel index
        mu    : float
           variance for Hirshfeld density
        dtype : dtype
           real or cmplx rotation matrix
        seed  : int
           seed for random initial guess for unitary matrix
        ----------

    """

    def __init__(self, wfs=None, calc=None,
                 method='W', penalty=2.0, spin=0,
                 mu=None, dtype=None, seed=None):
        from ase.dft.wannier import get_kklst, get_invkklst

        assert wfs or calc is not None

        if calc is not None:
            self.wfs = calc.wfs
        else:
            self.wfs = wfs  # CMOs

        if hasattr(self.wfs, 'mode'):
            self.mode = self.wfs.mode
        else:
            self.mode = None

        self.method = method  # Charge partitioning scheme
        self.penalty = abs(penalty)  # penalty exponent
        self.mu = mu  # WF variance (if 'H')

        self.gd = self.wfs.gd
        # Allow complex rotations
        if dtype is not None:
            self.dtype = dtype
        else:
            self.dtype = self.wfs.dtype

        self.setups = self.wfs.setups

        # Make atoms object from setups
        if calc is not None:
            self.atoms = calc.atoms
        else:
            self.atoms = get_atoms_object_from_wfs(self.wfs)

        self.Na = len(self.atoms)
        self.ns = self.wfs.nspins
        self.spin = spin
        self.niter = 0
        self.rng = np.random.default_rng(seed)

        # Determine nocc: integer occupations only
        k_rank, u = divmod(0 + len(self.wfs.kd.ibzk_kc) * spin,
                           len(self.wfs.kpt_u))

        f_n = self.wfs.kpt_u[u].f_n
        self.nocc = 0
        while f_n[self.nocc] > 1e-10:
            self.nocc += 1

        # Hold on to
        self.P = 0
        self.P_n = []
        self.Qa_nn = np.zeros((self.Na, self.nocc, self.nocc))

        # kpts and dirs
        self.k_kc = self.wfs.kd.bzk_kc

        assert len(self.wfs.kd.ibzk_kc) == len(self.k_kc)

        self.kgd = get_monkhorst_pack_size_and_offset(self.k_kc)[0]
        self.k_kc *= -1  # Bloch phase sign conv. GPAW

        # pbc-lattice
        self.Nk = len(self.k_kc)
        self.W_k = np.zeros((self.Nk, self.nocc, self.nocc),
                            dtype=self.dtype)

        # Expand cell to capture Bloch states
        largecell = (self.atoms.cell.T * self.kgd).T
        self.wd, self.Gd = calculate_weights(largecell)
        self.Nd = len(self.wd)

        # Get neighbor kpt list and inverse kpt list
        self.lst_dk, k0_dk = get_kklst(self.k_kc, self.Gd)
        self.invlst_dk = get_invkklst(self.lst_dk)

        # Using WFa and k-d lists make overlap matrix
        Qadk_nm = np.zeros((self.Na,
                            self.Nd,
                            self.Nk,
                            self.nocc, self.nocc), complex)

        if calc is not None and self.wfs.kpt_u[0].psit_nG is None:
            self.wfs.initialize_wave_functions_from_restart_file()

        # initialize wfs array if lcao
        if self.mode == 'lcao' and self.wfs.kpt_u[0].psit_nG is None:
            self.wfs.initialize_wave_functions_from_lcao()

        for d, dG in enumerate(self.Gd):
            for k in range(self.Nk):
                k1 = self.lst_dk[d, k]
                k0 = k0_dk[d, k]
                k_kc = self.wfs.kd.bzk_kc
                Gc = k_kc[k1] - k_kc[k] - k0
                # Det. kpt/spin
                kr, u = divmod(k + len(self.wfs.kd.ibzk_kc) * spin,
                               len(self.wfs.kpt_u))
                kr1, u1 = divmod(k1 + len(self.wfs.kd.ibzk_kc) * spin,
                                 len(self.wfs.kpt_u))

                if self.wfs.mode == 'pw':
                    cmo = self.gd.zeros(self.nocc, dtype=self.wfs.dtype)
                    cmo1 = self.gd.zeros(self.nocc, dtype=self.wfs.dtype)
                    for i in range(self.nocc):
                        cmo[i] = self.wfs._get_wave_function_array(u, i)
                        cmo1[i] = self.wfs._get_wave_function_array(u1, i)
                else:
                    cmo = self.wfs.kpt_u[u].psit_nG[:self.nocc]
                    cmo1 = self.wfs.kpt_u[u1].psit_nG[:self.nocc]
                # Inner product
                e_G = np.exp(-2j * pi *
                             np.dot(np.indices(self.gd.n_c).T +
                                    self.gd.beg_c,
                                    Gc / self.gd.N_c).T)
                # WFs per atom
                for atom in self.atoms:
                    WF = self.get_weight_function_atom(atom.index)
                    pw = (e_G * WF * cmo1)
                    Qadk_nm[atom.index, d, k] += \
                        self.gd.integrate(np.asarray(cmo, dtype=complex),
                                          pw,
                                          global_integral=False)
                # PAW corrections
                P_ani1 = self.wfs.kpt_u[u1].P_ani

                spos_ac = self.atoms.get_scaled_positions()

                for A, P_ni in self.wfs.kpt_u[u].P_ani.items():
                    dS_ii = self.setups[A].dO_ii
                    P_n = P_ni[:self.nocc]
                    P_n1 = P_ani1[A][:self.nocc]
                    # Phase factor is an approx. PRB 72, 125119 (2005)
                    e = np.exp(-2j * pi * np.dot(Gc, spos_ac[A]))
                    Qadk_nm[A, d, k] += \
                        e * P_n.conj().dot(dS_ii.dot(P_n1.T))

        # Sum over domains
        self.gd.comm.sum(Qadk_nm)
        self.Qadk_nm = Qadk_nm.copy()
        self.Qadk_nn = np.zeros_like(self.Qadk_nm)

        # Initial W_k: Start from random WW*=I
        for k in range(self.Nk):
            self.W_k[k] = random_orthogonal(self.rng, self.nocc,
                                            dtype=self.dtype)
        if world is not None:
            world.broadcast(self.W_k, 0)

        # Given all matrices, update
        self.update()
        self.initialized = True

    def step(self, dX):
        No = self.nocc
        Nk = self.Nk

        A_kww = dX[:Nk * No ** 2].reshape(Nk, No, No)
        for U, A in zip(self.W_k, A_kww):
            H = -1.j * A.conj()
            epsilon, Z = np.linalg.eigh(H)
            dU = np.dot(Z * np.exp(1.j * epsilon), dagger(Z))
            if U.dtype == float:
                U[:] = np.dot(U, dU).real
            else:
                U[:] = np.dot(U, dU)
        self.update()

    def get_weight_function_atom(self, index):
        if self.method == 'H':
            WFa = WeightFunc(self.gd,
                             self.atoms,
                             [index],
                             mu=self.mu
                             ).construct_weight_function()
        elif self.method == 'W':
            WFa = WignerSeitz(self.gd,
                              self.atoms,
                              index
                              ).construct_weight_function()
        else:
            raise ValueError('check method')
        return WFa

    def localize(self, step=0.25, tolerance=1e-8, verbose=False):
        md_min(self, step, tolerance, verbose)

    def update(self):
        for a in range(self.Na):
            for d in range(self.Nd):
                for k in range(self.Nk):
                    k1 = self.lst_dk[d, k]
                    self.Qadk_nn[a, d, k] = \
                        np.dot(self.W_k[k].T.conj(),
                               np.dot(self.Qadk_nm[a, d, k],
                                      self.W_k[k1]))
        # Update PCM
        self.Qad_nn = self.Qadk_nn.sum(axis=2) / self.Nk

    def update_matrices(self):
        # Using new W_k rotate states
        for a in range(self.Na):
            for d in range(self.Nd):
                for k in range(self.Nk):
                    k1 = self.lst_dk[d, k]
                    self.Qadk_nn[a, d, k] = \
                        np.dot(self.W_k[k].T.conj(),
                               np.dot(self.Qadk_nm[a, d, k],
                                      self.W_k[k1]))

    def get_function_value(self):
        # Over k
        Qad_nn = np.sum(abs(self.Qadk_nn), axis=2) / self.Nk
        # Over d
        Qa_nn = 0
        self.P = 0
        for d in range(self.Nd):
            Qa_nn += Qad_nn[:, d] ** 2 * self.wd[d]
        # Over a and diag
        for a in range(self.Na):
            self.P += np.sum(Qa_nn[a].diagonal())

        self.P /= np.sum(self.wd)
        self.P_n.append(self.P)

        return self.P

    def get_gradients(self):
        No = self.nocc
        dW = []

        for k in range(self.Nk):
            Wtemp = np.zeros((No, No), complex)

            for a in range(self.Na):
                for d, wd in enumerate(self.wd):
                    diagQ = self.Qad_nn[a, d].diagonal()
                    Qa_ii = np.repeat(diagQ, No).reshape(No, No)
                    k2 = self.invlst_dk[d, k]
                    Qk_nn = self.Qadk_nn[a, d]
                    temp = Qa_ii.T * Qk_nn[k].conj() - \
                        Qa_ii * Qk_nn[k2].conj()
                    Wtemp += wd * (temp - dagger(temp))

            dW.append(Wtemp.ravel())

        return np.concatenate(dW)



"""
A class for finding optimal orbitals corresponding to a stationary point of
the energy functional using direct optimization and exponential transformation
in LCAO mode.

It can be used with Kohn-Sham functionals and can include Perdew-Zunger
self-interaction correction (PZ-SIC) in the calculations.

Ground state as well as variational excited state calculations can be
performed. Ground state calculations involve minimization of the energy
(direct minimization), while excited state calculations involve convergence
on a saddle point (direct optimization).

Implementation of exponential transformation direct minimization (ETDM) for
ground state calculations LCAO mode:

    Comput. Phys. Commun. 267, 108047 (2021) :doi:10.1016/j.cpc.2021.108047
    arXiv:2101.12597 [physics.comp-ph]

GPAW implementations of direct optimization (DO) for variational excited state
calculations LCAO mode:

    DO with preconditioned quasi-Newton algorithms and maximum overlap method
    (DO-MOM)
    J. Chem. Theory Comput. 16, 6968 (2020) :doi:10.1021/acs.jctc.0c00597
    arXiv:2006.15922 [physics.chem-ph]

    DO with generalized mode following (DO-GMF)
    J. Chem. Theory Comput. 19, 3634 (2023) :doi:10.1021/acs.jctc.3c00178
    arXiv:2302.05912 [physics.chem-ph]
"""


import numpy as np
import warnings
from ase.utils import basestring
from gpaw.directmin.tools import expm_ed, expm_ed_unit_inv, random_a, \
    sort_orbitals_according_to_occ, sort_orbitals_according_to_energies
from gpaw.directmin.lcao.etdm_helper_lcao import ETDMHelperLCAO
from gpaw.directmin.locfunc.localize_orbitals import localize_orbitals
from scipy.linalg import expm
from gpaw.directmin import search_direction, line_search_algorithm
from gpaw.directmin.tools import get_n_occ
from gpaw.directmin.derivatives import get_approx_analytical_hessian
from gpaw import BadParallelization
from copy import deepcopy


class LCAOETDM:

    def __init__(self,
                 excited_state=False,
                 searchdir_algo=None,
                 linesearch_algo=None,
                 partial_diagonalizer='Davidson',
                 update_ref_orbs_counter=20,
                 update_ref_orbs_canonical=False,
                 update_precond_counter=1000,
                 use_prec=True,
                 matrix_exp='pade-approx',
                 representation='sparse',
                 functional='ks',
                 need_init_orbs=None,
                 orthonormalization='gramschmidt',
                 randomizeorbitals=None,
                 checkgraderror=False,
                 need_localization=True,
                 localizationtype=None,
                 localizationseed=None,
                 constraints=None,
                 subspace_convergence=5e-4
                 ):
        """Class for direct orbital optimization in LCAO mode.

        Parameters
        ----------
        excited_state: bool
            If False (default), use search direction and line search
            algorithms for ground state calculation ('l-bfgs-p' and 'swc-awc')
            if not specified by searchdir_algo linesearch_algo, and set
            need_init_orbs to True. Otherwise, use search direction and line
            search algorithms for excited state calculation ('l-sr1p' and
            'max-step'), and set need_init_orbs to False.
        searchdir_algo: str, dict or instance
            Search direction algorithm. Can be one of the algorithms available
            in sd_etdm.py:
                'sd': Steepest descent
                'fr-cg': Fletcher-Reeves conjugate gradient
                'l-bfgs': Limited-memory BFGS
                'l-bfgs-p': Limited-memory BFGS with preconditioner presented
                    in :doi:`10.1016/j.cpc.2021.108047` (default when
                    excited_state is False)
                'l-sr1p': limited-memory SR1 algorithm presented in
                    :doi:`10.1021/acs.jctc.0c00597` (default when excited_state
                    is True)
            The default memory for 'l-bfgs'/'l-bfgs-p' and 'l-sr1p' is 3 and
            20, respectively, and can be changed by supplying a dictionary:
            {'name': name, 'memory': memory}, where name should be 'l-bfgs',
            'l-bfgs-p' or 'l-sr1p' and memory should be an int.
            To use the generalized mode following (GMF) method for excited
            states, append '_gmf' to the search direction algorithm name. E.g.
            'l-bfgs-p_gmf'.
        linesearch_algo: str, dict or instance
            Line search algorithm. Can be one of the algorithms available
            in ls_etdm.py:
                'max-step': The quasi-Newton step is scaled if it exceeds a
                    maximum step length (default when excited_state is True).
                    The default maximum step length is 0.20, and can be changed
                    by supplying a dictionary:
                    {'name': 'max-step', 'max_step': max_step}, where max_step
                    should be a float
                'swc-awc': Line search with Wolfe conditions (default when
                    excited_state is False)
        partial_diagonalizer: 'str' or dict
            Algorithm for partial diagonalization of the electronic Hessian if
            GMF is used. Default is 'Davidson'.
        update_ref_orbs_counter: int
            When to update the coefficients of the reference orbitals. Default
            is 20 iterations.
        update_ref_orbs_canonical: bool
            If True, the coefficients of the reference orbitals are updated to
            canonical orbital coefficients, otherwise use the optimal orbital
            coefficients (default).
        use_prec: bool
            If True (default) use a preconditioner. The preconditioner is
            calculated as the inverse of a diagonal approximation of the
            Hessian (see :doi:`10.1021/j100322a012`) apart for 'l-bfgs-p',
            which uses the composite preconditioner presented in
            :doi:`10.1016/j.cpc.2021.108047`.
        update_precond_counter: int
            When to update the preconditioner. Default is 1000 iterations.
        representation: 'str'
            How to store the elements of the anti-Hermitian matrix for the
            matrix exponential. Can be one of 'sparse' (default), 'full',
            'u-invar'. The latter can be used only for unitary invariant
            functionals, such as Kohn-Sham functionals when the occupied
            orbitals have the same occupation number, but not for orbital
            density dependent functionals, such as when using PZ-SIC.
        matrix_exp: 'str'
            Algorithm for calculating the matrix exponential and the gradient
            with respect to the elements of the anti-Hermitian matrix for the
            exponential transformation. Can be one of 'pade-approx' (default),
            'egdecomp', 'egdecomp-u-invar' (the latter can be used only with
            'u-invar' representation).
        functional: str, dict or instance
            Type of functional. Can be one of:
                'ks': The functional as specified in the GPAW calculator is
                    used (default)
                'pz-sic': Apply the Perdew-Zunger self-interaction correction
                    on top of the functional as specified in the GPAW
                    calculator. Dy default full SIC is applied. A scaling
                    factor for SIC can be given by supplying a dictionary:
                    functional={'name': 'pz-sic', 'scaling_factor': (a, a)},
                    where a is the scaling factor (float).
        need_init_orbs: bool
            If True (default when excited_state is False), obtain initial
            orbitals from eigendecomposition of the Hamiltonian matrix. If
            False (default when excited_state is True), use orbitals stored in
            wfs object as initial guess.
        orthonormalization: str
            Method to orthonormalize the orbitals. Can be one of 'gramschmidt'
            (Gram-Schmidt orthonormalization, default), 'loewdin' (Loewdin
            orthonormalization) or 'diag' (eigendecomposition of the
            Hamiltonian matrix).
        randomizeorbitals: numpy.random.Generator
            Optional RNG to add random noise to the initial guess orbitals,
            default is to not add noise to them
        checkgraderror: bool
            If True, can be used to check the error in the estimation of the
            gradient (only with representation 'full'). Default is False.
        need_localization: bool
            If True (default), localize initial guess orbitals. Requires a
            specification of localizationtype, otherwise it is set to False.
            Recommended for calculations with PZ-SIC.
        localizationtype: str
            Method for localizing the initial guess orbitals. Can be one of:
                'pz': Unitary optimization among occupied orbitals (subspace
                    optimization) with PZ-SIC
                'pm': Pipek-Mezey localization (recommended for PZ-SIC)
                'fb' Foster-Boys localization
            Default is None, meaning that no localization is performed.
        localizationseed: int
            Seed for Pipek-Mezey or Foster-Boys localization. Default is
            None (no seed is used).
        constraints: list of lists of int
            List of constraints on the orbital rotation for each k-point. If
            a constraint is given as a pair of orbital indices, the rotation
            between these two orbitals is constrained. If a constraint is
            given as a single index, the orbital with this index is frozen.
            E.g.: calc.set(eigensolver=LCAOETDM(constraints=[[[i]], [[j]]]))
            will freeze orbital i in the first k-point and orbital j in the
            second k-point.
        subspace_convergence: float
            Tolerance on the norm of the gradient for convergence of the
            subspace optimization with PZ-SIC.
        """

        assert representation in ['sparse', 'u-invar', 'full'], 'Value Error'
        assert matrix_exp in ['egdecomp', 'egdecomp-u-invar', 'pade-approx'], \
            'Value Error'
        if matrix_exp == 'egdecomp-u-invar':
            assert representation == 'u-invar', 'Use u-invar representation ' \
                                                'with egdecomp-u-invar'
        assert orthonormalization in ['gramschmidt', 'loewdin', 'diag'], \
            'Value Error'

        self.sda = searchdir_algo
        self.lsa = linesearch_algo
        self.partial_diagonalizer = partial_diagonalizer
        self.localizationseed = localizationseed
        self.update_ref_orbs_counter = update_ref_orbs_counter
        self.update_ref_orbs_canonical = update_ref_orbs_canonical
        self.update_precond_counter = update_precond_counter
        self.use_prec = use_prec
        self.matrix_exp = matrix_exp
        self.localizationtype = localizationtype
        self.need_localization = need_localization
        self.need_init_orbs = need_init_orbs
        self.randomizeorbitals = randomizeorbitals
        self.representation = representation
        self.orthonormalization = orthonormalization
        self.constraints = constraints
        self.subspace_convergence = subspace_convergence
        self.released_subspace = False
        self.excited_state = excited_state
        self.name = 'etdm-lcao'
        self.iters = 0
        self.eg_count = 0
        self.subspace_iters = 0
        self.restart = False
        self.gmf = False
        self.check_inputs_and_init_search_algo()

        self.checkgraderror = checkgraderror
        self._norm_commutator, self._norm_grad = 0., 0.
        self.error = 0
        self.e_sic = 0.0
        self.subspace_optimization = False

        # these are things we cannot initialize now
        self.func_settings = functional
        self.dtype = None
        self.nkpts = None
        self.gd = None
        self.nbands = None

        # values: vectors of the elements of matrices, keys: kpt number
        self.a_vec_u = {}  # for the elements of the skew-Hermitian matrix A
        self.a_vec_oo_u = {}
        self.a_vec_ov_u = {}
        self.a_vec_all_u = {}
        self.g_vec_u = {}  # for the elements of the gradient matrix G
        self.g_vec_oo_u = {}
        self.g_vec_ov_u = {}
        self.g_vec_all_u = {}
        self.evecs = {}   # eigenvectors for i*a_vec_u
        self.evals = {}   # eigenvalues for i*a_vec_u
        self.ind_up = {}
        self.ind_oo_up = {}
        self.ind_ov_up = {}
        self.ind_sparse_up = {}
        self.ind_all_up = {}
        self.n_dim = {}
        self.n_dim_oo = {}
        self.n_dim_all = {}
        self.alpha = 1.0  # step length
        self.phi_2i = [None, None]  # energy at last two iterations
        self.der_phi_2i = [None, None]  # energy gradient w.r.t. alpha
        self.hess = {}  # approximate Hessian

        # for mom
        self.initial_occupation_numbers = None

        # in this attribute we store the object specific to each mode
        self.dm_helper = None

        self.initialized = False

    def check_inputs_and_init_search_algo(self):
        defaults = self.set_defaults()
        if self.sda is None:
            self.sda = defaults['searchdir_algo']
        if self.lsa is None:
            self.lsa = defaults['linesearch_algo']
        if self.need_init_orbs is None:
            self.need_init_orbs = defaults['need_init_orbs']
        if self.localizationtype is None:
            self.need_localization = False

        self.searchdir_algo = search_direction(
            self.sda, self, self.partial_diagonalizer)
        sd_name = self.searchdir_algo.name.split('_')
        if sd_name[0] == 'l-bfgs-p' and not self.use_prec:
            raise ValueError('Use l-bfgs-p with use_prec=True')
        if len(sd_name) == 2:
            if sd_name[1] == 'gmf':
                self.searchdir_algo.name = sd_name[0]
                self.gmf = True
                self.g_vec_u_original = None
                self.pd = self.partial_diagonalizer

        self.line_search = line_search_algorithm(self.lsa,
                                                 self.evaluate_phi_and_der_phi,
                                                 self.searchdir_algo)

    def set_defaults(self):
        if self.excited_state:
            return {'searchdir_algo': 'l-sr1p',
                    'linesearch_algo': 'max-step',
                    'need_init_orbs': False}
        else:
            return {'searchdir_algo': 'l-bfgs-p',
                    'linesearch_algo': 'swc-awc',
                    'need_init_orbs': True}

    def __repr__(self):

        sda_name = self.searchdir_algo.name
        lsa_name = self.line_search.name
        if isinstance(self.func_settings, basestring):
            func_name = self.func_settings
        else:
            func_name = self.func_settings['name']
        if self.gmf:
            if isinstance(self.pd, basestring):
                pd_name = self.pd
            else:
                pd_name = self.pd['name']

        add = ''
        pd_add = ''
        if self.gmf:
            add = ' with minimum mode following'
            pardi = {'Davidson': 'Finite difference generalized Davidson '
                     'algorithm'}
            pd_add = '       ' \
                     'Partial diagonalizer: {}\n'.format(
                         pardi[pd_name])

        sds = {'sd': 'Steepest Descent',
               'fr-cg': 'Fletcher-Reeves conj. grad. method',
               'l-bfgs': 'L-BFGS algorithm',
               'l-bfgs-p': 'L-BFGS algorithm with preconditioning',
               'l-sr1p': 'Limited-memory SR1P algorithm'}

        lss = {'max-step': 'step size equals one',
               'swc-awc': 'Inexact line search based on cubic interpolation,\n'
                          '                    strong and approximate Wolfe '
                          'conditions'}

        repr_string = 'Direct minimisation' + add + ' using exponential ' \
                      'transformation.\n'
        repr_string += '       ' \
                       'Search ' \
                       'direction: {}\n'.format(sds[sda_name] + add)
        repr_string += pd_add
        repr_string += '       ' \
                       'Line ' \
                       'search: {}\n'.format(lss[lsa_name])
        repr_string += '       ' \
                       'Preconditioning: {}\n'.format(self.use_prec)
        repr_string += '       ' \
                       'Orbital-density self-interaction ' \
                       'corrections: {}\n'.format(func_name)
        repr_string += '       ' \
                       'WARNING: do not use it for metals as ' \
                       'occupation numbers are\n' \
                       '                ' \
                       'not found variationally\n'

        return repr_string

    def initialize(self, gd, dtype, nbands, nkpts, nao, using_blacs,
                   bd_comm_size, kpt_u):

        assert nbands == nao, \
            'Please, use: nbands=\'nao\''
        if not bd_comm_size == 1:
            raise BadParallelization(
                'Band parallelization is not supported')
        if using_blacs:
            raise BadParallelization(
                'ScaLapack parallelization is not supported')

        self.gd = gd
        self.dtype = dtype
        self.nbands = nbands
        self.nkpts = nkpts

        for kpt in kpt_u:
            u = self.kpointval(kpt)
            # dimensionality of the problem.
            # this implementation rotates among all bands
            self.n_dim_all[u] = self.nbands

        self.initialized = True

        # need reference orbitals
        self.dm_helper = None

    def initialize_dm_helper(self, wfs, ham, dens, log):

        self.dm_helper = ETDMHelperLCAO(
            wfs, dens, ham, self.nkpts, self.func_settings,
            diagonalizer=None,
            orthonormalization=self.orthonormalization,
            need_init_orbs=self.need_init_orbs)
        self.need_init_orbs = self.dm_helper.need_init_orbs

        # randomize orbitals?
        if self.randomizeorbitals is not None:
            for kpt in wfs.kpt_u:
                self.randomize_orbitals_kpt(wfs, kpt)
            self.randomizeorbitals = None

        wfs.calculate_occupation_numbers(dens.fixed)

        # MOM
        self.initial_sort_orbitals_mom(wfs)

        # initialize matrices
        self.set_variable_matrices(wfs.kpt_u)

        # localize orbitals?
        self.localize(wfs, dens, ham, log)

        # MOM
        self.initialize_mom_reference_orbitals(wfs, dens)

        for kpt in wfs.kpt_u:
            f_unique = np.unique(kpt.f_n)
            if len(f_unique) > 2 and self.representation == 'u-invar':
                warnings.warn("Use representation == 'sparse' when "
                              "there are unequally occupied orbitals "
                              "as the functional is not unitary invariant")

        # set reference orbitals
        self.dm_helper.set_reference_orbitals(wfs, self.n_dim)

    def set_variable_matrices(self, kpt_u):

        # Matrices are sparse and Skew-Hermitian.
        # They have this structure:
        #  A_BigMatrix =
        #
        # (  A_1          A_2 )
        # ( -A_2.T.conj() 0   )
        #
        # where 0 is a zero-matrix of size of (M-N) * (M-N)
        #
        # A_1 i skew-hermitian matrix of N * N,
        # N-number of occupied states
        # A_2 is matrix of size of (M-N) * N,
        # M - number of basis functions
        #
        # if the energy functional is unitary invariant
        # then A_1 = 0
        # (see Hutter J et. al, J. Chem. Phys. 101, 3862 (1994))
        #
        # We will keep A_1 as we would like to work with metals,
        # SIC, and molecules with different occupation numbers.
        # this corresponds to 'sparse' representation
        #
        # Thus, for the 'sparse' we need to store upper
        # triangular part of A_1, and matrix A_2, so in total
        # (M-N) * N + N * (N - 1)/2 = N * (M - (N + 1)/2) elements
        #
        # we will store these elements as a vector and
        # also will store indices of the A_BigMatrix
        # which correspond to these elements.
        #
        # 'u-invar' corresponds to the case when we want to
        # store only A_2, that is this representation is sparser

        for kpt in kpt_u:
            n_occ = get_n_occ(kpt)[0]
            u = self.kpointval(kpt)
            self.n_dim[u] = deepcopy(self.n_dim_all[u])
            self.n_dim_oo[u] = n_occ
            # M - one dimension of the A_BigMatrix
            M = self.n_dim_all[u]
            i1_oo, i2_oo = [], []
            for i in range(n_occ):
                for j in range(i + 1, n_occ):
                    i1_oo.append(i)
                    i2_oo.append(j)
            self.ind_oo_up[u] = (np.asarray(i1_oo), np.asarray(i2_oo))
            i1_ov, i2_ov = [], []
            for i in range(n_occ):
                for j in range(n_occ, M):
                    i1_ov.append(i)
                    i2_ov.append(j)
            self.ind_ov_up[u] = (np.asarray(i1_ov), np.asarray(i2_ov))
            i1_sparse, i2_sparse = [], []
            for i in range(n_occ):
                for j in range(i + 1, M):
                    i1_sparse.append(i)
                    i2_sparse.append(j)
            self.ind_sparse_up[u] = (np.asarray(i1_sparse),
                                     np.asarray(i2_sparse))
            if self.representation == 'u-invar':
                self.ind_all_up[u] = self.ind_ov_up[u]
            if self.representation == 'sparse':
                self.ind_all_up[u] = self.ind_sparse_up[u]
            elif self.representation == 'full' and self.dtype == complex:
                # Take indices of all upper triangular and diagonal
                # elements of A_BigMatrix
                self.ind_all_up[u] = np.triu_indices(self.n_dim[u])

            shape_of_oo = len(self.ind_oo_up[u][0])
            shape_of_ov = len(self.ind_ov_up[u][0])
            shape_of_all = len(self.ind_all_up[u][0])

            self.a_vec_oo_u[u] = np.zeros(shape=shape_of_oo, dtype=self.dtype)
            self.a_vec_ov_u[u] = np.zeros(shape=shape_of_ov, dtype=self.dtype)
            self.a_vec_all_u[u] = np.zeros(
                shape=shape_of_all, dtype=self.dtype)
            self.g_vec_oo_u[u] = np.zeros(shape=shape_of_oo, dtype=self.dtype)
            self.g_vec_ov_u[u] = np.zeros(shape=shape_of_ov, dtype=self.dtype)
            self.g_vec_all_u[u] = np.zeros(
                shape=shape_of_all, dtype=self.dtype)
            self.evecs[u] = None
            self.evals[u] = None

            # All constraints passed as a list of a single orbital index are
            # converted to all lists of orbital pairs involving that orbital
            # index
            if self.constraints:
                self.constraints[u] = convert_constraints(
                    self.constraints[u], self.n_dim[u],
                    len(kpt.f_n[kpt.f_n > 1e-10]), self.representation)

            # If there are no degrees of freedom no need to optimize.
            # Indicated by setting n_dim to 0, as n_dim can never be 0
            # otherwise.
            for k in self.ind_all_up:
                if not self.ind_all_up[k][0].size \
                        or not self.ind_all_up[k][1].size:
                    self.n_dim_all[k] = 0  # Skip full space optimization
                if not self.ind_oo_up[k][0].size \
                        or not self.ind_oo_up[k][1].size:
                    self.n_dim_oo[k] = 0  # Skip PZ localization if requested

        self.ind_up = deepcopy(self.ind_all_up)
        self.a_vec_u = deepcopy(self.a_vec_all_u)
        self.g_vec_u = deepcopy(self.g_vec_all_u)

        # This conversion makes it so that constraint-related functions can
        # iterate through a list of no constraints rather than checking for
        # None every time
        if self.constraints is None:
            self.constraints = [[] for _ in range(len(kpt_u))]

        self.iters = 1

    def localize(self, wfs, dens, ham, log):
        if self.need_localization:
            localizationtype = \
                self.localizationtype.replace('-', '').lower().split('_')
            do_oo_subspace = 'pz' in localizationtype
            localize_orbitals(
                wfs, dens, ham, log, self.localizationtype,
                seed=self.localizationseed)
            if do_oo_subspace:
                assert self.dm_helper.func.name == 'PZ-SIC', \
                    'PZ-SIC localization requested, but functional ' \
                    'settings do not use PZ-SIC.'
                self.lock_subspace('oo')
            self.need_localization = False

    def lock_subspace(self, subspace='oo'):
        self.subspace_optimization = True
        self.subspace_iters = 1
        if subspace == 'oo':
            self.n_dim = deepcopy(self.n_dim_oo)
            self.ind_up = deepcopy(self.ind_oo_up)
            self.a_vec_u = deepcopy(self.a_vec_oo_u)
            self.g_vec_u = deepcopy(self.g_vec_oo_u)
        elif subspace == 'ov':
            self.n_dim = deepcopy(self.n_dim_all)
            self.ind_up = deepcopy(self.ind_ov_up)
            self.a_vec_u = deepcopy(self.a_vec_ov_u)
            self.g_vec_u = deepcopy(self.g_vec_ov_u)
        self.alpha = 1.0
        self.phi_2i = [None, None]
        self.der_phi_2i = [None, None]

    def release_subspace(self):
        self.subspace_optimization = False
        self.released_subspace = True
        self.n_dim = deepcopy(self.n_dim_all)
        self.ind_up = deepcopy(self.ind_all_up)
        self.a_vec_u = deepcopy(self.a_vec_all_u)
        self.g_vec_u = deepcopy(self.g_vec_all_u)
        self.alpha = 1.0
        self.phi_2i = [None, None]
        self.der_phi_2i = [None, None]

    def iterate(self, ham, wfs, dens):
        """
        One iteration of direct optimization
        for occupied orbitals

        :param ham:
        :param wfs:
        :param dens:
        :return:
        """
        with wfs.timer('Direct Minimisation step'):
            self.update_ref_orbitals(wfs, ham, dens)

            a_vec_u = self.a_vec_u
            alpha = self.alpha
            phi_2i = self.phi_2i
            der_phi_2i = self.der_phi_2i
            c_ref = self.dm_helper.reference_orbitals

            if self.iters == 1 or self.released_subspace or \
                    (self.subspace_optimization and self.subspace_iters == 1):
                phi_2i[0], g_vec_u = \
                    self.get_energy_and_gradients(
                        a_vec_u, ham, wfs, dens, c_ref)
            else:
                g_vec_u = self.g_vec_u_original if self.gmf \
                    and not self.subspace_optimization else self.g_vec_u

            make_pd = False
            if self.gmf and not self.subspace_optimization:
                with wfs.timer('Partial Hessian diagonalization'):
                    self.searchdir_algo.update_eigenpairs(
                        g_vec_u, wfs, ham, dens)
                # The diagonal Hessian approximation must be positive-definite
                make_pd = True

            with wfs.timer('Preconditioning:'):
                precond = self.get_preconditioning(
                    wfs, self.use_prec, make_pd=make_pd)

            with wfs.timer('Get Search Direction'):
                # calculate search direction according to chosen
                # optimization algorithm (e.g. L-BFGS)
                p_vec_u = self.searchdir_algo.update_data(
                    wfs, a_vec_u, g_vec_u, precond=precond,
                    subspace=self.subspace_optimization)

            # recalculate derivative with new search direction
            der_phi_2i[0] = 0.0
            for k in g_vec_u:
                der_phi_2i[0] += g_vec_u[k].conj() @ p_vec_u[k]
            der_phi_2i[0] = der_phi_2i[0].real
            der_phi_2i[0] = wfs.kd.comm.sum_scalar(der_phi_2i[0])

            alpha, phi_alpha, der_phi_alpha, g_vec_u = \
                self.line_search.step_length_update(
                    a_vec_u, p_vec_u, wfs, ham, dens, c_ref, phi_0=phi_2i[0],
                    der_phi_0=der_phi_2i[0], phi_old=phi_2i[1],
                    der_phi_old=der_phi_2i[1], alpha_max=5.0, alpha_old=alpha,
                    kpdescr=wfs.kd)

            if wfs.gd.comm.size > 1:
                with wfs.timer('Broadcast gradients'):
                    alpha_phi_der_phi = np.array([alpha, phi_2i[0],
                                                  der_phi_2i[0]])
                    wfs.gd.comm.broadcast(alpha_phi_der_phi, 0)
                    alpha = alpha_phi_der_phi[0]
                    phi_2i[0] = alpha_phi_der_phi[1]
                    der_phi_2i[0] = alpha_phi_der_phi[2]
                    for kpt in wfs.kpt_u:
                        k = self.kpointval(kpt)
                        wfs.gd.comm.broadcast(g_vec_u[k], 0)

            # calculate new matrices for optimal step length
            for k in a_vec_u:
                a_vec_u[k] += alpha * p_vec_u[k]
            self.alpha = alpha
            self.g_vec_u = g_vec_u
            self.iters += 1

            # and 'shift' phi, der_phi for the next iteration
            phi_2i[1], der_phi_2i[1] = phi_2i[0], der_phi_2i[0]
            phi_2i[0], der_phi_2i[0] = phi_alpha, der_phi_alpha,

            if self.subspace_optimization:
                if self.get_grad_norm() < self.subspace_convergence:
                    self.release_subspace()
                    self.dm_helper.set_reference_orbitals(wfs, self.n_dim)
                    self.searchdir_algo.reset()
                    for k, kpt in enumerate(wfs.kpt_u):
                        self.hess[k] = get_approx_analytical_hessian(
                            kpt, self.dtype, ind_up=self.ind_up[k])
                        wfs.atomic_correction.calculate_projections(wfs, kpt)
                self.error = np.inf  # Do not consider this converged!

    def get_grad_norm(self):
        norm = 0.0
        for k in self.g_vec_u.keys():
            norm += np.linalg.norm(self.g_vec_u[k])
        return norm

    def get_energy_and_gradients(self, a_vec_u, ham, wfs, dens,
                                 c_ref):

        """
        Energy E = E[C_ref exp(A)]. Gradients G_ij[C, A] = dE/dA_ij

        :param wfs:
        :param ham:
        :param dens:
        :param a_vec_u: A
        :param c_ref: C_ref
        :return:
        """

        self.rotate_wavefunctions(wfs, a_vec_u, c_ref)

        e_total = self.update_ks_energy(ham, wfs, dens)

        with wfs.timer('Calculate gradients'):
            g_vec_u = {}
            self.error = 0.0
            self.e_sic = 0.0  # this is odd energy
            for kpt in wfs.kpt_u:
                k = self.kpointval(kpt)
                if self.n_dim[k] == 0:
                    g_vec_u[k] = np.zeros_like(a_vec_u[k])
                    continue
                g_vec_u[k], error = self.dm_helper.calc_grad(
                    wfs, ham, kpt, self.evecs[k], self.evals[k],
                    self.matrix_exp, self.representation, self.ind_up[k],
                    self.constraints[k])

                if hasattr(self.dm_helper.func, 'e_sic_by_orbitals'):
                    self.e_sic \
                        += self.dm_helper.func.e_sic_by_orbitals[k].sum()

                self.error += error
            self.error = wfs.kd.comm.sum_scalar(self.error)
            self.e_sic = wfs.kd.comm.sum_scalar(self.e_sic)

        self.eg_count += 1

        if self.representation == 'full' and self.checkgraderror:
            self._norm_commutator = 0.0
            for kpt in wfs.kpt_u:
                u = self.kpointval(kpt)
                a_mat = vec2skewmat(a_vec_u[u], self.n_dim[u],
                                    self.ind_up[u], wfs.dtype)
                g_mat = vec2skewmat(g_vec_u[u], self.n_dim[u],
                                    self.ind_up[u], wfs.dtype)

                tmp = np.linalg.norm(g_mat @ a_mat - a_mat @ g_mat)
                if self._norm_commutator < tmp:
                    self._norm_commutator = tmp

                tmp = np.linalg.norm(g_mat)
                if self._norm_grad < tmp:
                    self._norm_grad = tmp

        return e_total + self.e_sic, g_vec_u

    def update_ks_energy(self, ham, wfs, dens):
        """
        Update Kohn-Sham energy
        It assumes the temperature is zero K.
        """

        dens.update(wfs)
        ham.update(dens, wfs, False)

        return ham.get_energy(0.0, wfs, False)

    def evaluate_phi_and_der_phi(self, a_vec_u, p_mat_u, alpha,
                                 wfs, ham, dens, c_ref,
                                 phi=None, g_vec_u=None):
        """
        phi = f(x_k + alpha_k*p_k)
        der_phi = \\grad f(x_k + alpha_k*p_k) \\cdot p_k
        :return:  phi, der_phi # floats
        """
        if phi is None or g_vec_u is None:
            x_mat_u = {k: a_vec_u[k] + alpha * p_mat_u[k] for k in a_vec_u}
            phi, g_vec_u = self.get_energy_and_gradients(
                x_mat_u, ham, wfs, dens, c_ref)

            # If GMF is used save the original gradient and invert the parallel
            # projection onto the eigenvectors with negative eigenvalues
            if self.gmf and not self.subspace_optimization:
                self.g_vec_u_original = deepcopy(g_vec_u)
                g_vec_u = self.searchdir_algo.invert_parallel_grad(g_vec_u)

        der_phi = 0.0
        for k in p_mat_u:
            der_phi += g_vec_u[k].conj() @ p_mat_u[k]

        der_phi = der_phi.real
        der_phi = wfs.kd.comm.sum_scalar(der_phi)

        return phi, der_phi, g_vec_u

    def update_ref_orbitals(self, wfs, ham, dens):
        """
        Update reference orbitals

        :param wfs:
        :param ham:
        :return:
        """

        if self.representation == 'full':
            badgrad = self._norm_commutator > self._norm_grad / 3. and \
                self.checkgraderror
        else:
            badgrad = False
        counter = self.update_ref_orbs_counter
        if (self.iters % counter == 0 and self.iters > 1) or \
                (self.restart and self.iters > 1) or badgrad:
            self.iters = 1
            if self.update_ref_orbs_canonical or self.restart:
                self.get_canonical_representation(ham, wfs, dens)
            else:
                self.set_ref_orbitals_and_a_vec(wfs)

            # Erase memory of search direction algorithm
            self.searchdir_algo.reset()

    def get_preconditioning(self, wfs, use_prec, make_pd=False):

        if not use_prec:
            return None

        if self.searchdir_algo.name == 'l-bfgs-p':
            beta0 = self.searchdir_algo.beta_0
            gamma = 0.25
        else:
            beta0 = 1.0
            gamma = 0.0

        counter = self.update_precond_counter
        precond = {}
        for kpt in wfs.kpt_u:
            k = self.kpointval(kpt)
            w = kpt.weight / (3.0 - wfs.nspins)
            if self.iters % counter == 0 or self.iters == 1:
                self.hess[k] = get_approx_analytical_hessian(
                    kpt, self.dtype, ind_up=self.ind_up[k])
                if make_pd:
                    if self.dtype == float:
                        self.hess[k] = np.abs(self.hess[k])
                    else:
                        self.hess[k] = np.abs(self.hess[k].real) \
                            + 1.0j * np.abs(self.hess[k].imag)
            hess = self.hess[k]
            precond[k] = np.zeros_like(hess)
            correction = w * gamma * beta0 ** (-1)
            if self.searchdir_algo.name != 'l-bfgs-p':
                correction = np.zeros_like(hess)
                zeros = abs(hess) < 1.0e-4
                correction[zeros] = 1.0
            precond[k] += 1. / ((1 - gamma) * hess.real + correction)
            if self.dtype == complex:
                precond[k] += 1.j / ((1 - gamma) * hess.imag + correction)

        return precond

    def get_canonical_representation(self, ham, wfs, dens,
                                     sort_eigenvalues=False):
        """
        Choose canonical orbitals as the orbitals that diagonalize the
        Lagrange matrix. It is probably necessary to do a subspace rotation
        with equally occupied orbitals as the total energy is unitary invariant
        within equally occupied subspaces.
        """

        with ((wfs.timer('Get canonical representation'))):
            for kpt in wfs.kpt_u:
                self.dm_helper.update_to_canonical_orbitals(
                    wfs, ham, kpt, self.update_ref_orbs_canonical,
                    self.restart)

            if self.update_ref_orbs_canonical or self.restart:
                wfs.calculate_occupation_numbers(dens.fixed)
                sort_orbitals_according_to_occ(
                    wfs, self.constraints, update_mom=True)

            if sort_eigenvalues:
                sort_orbitals_according_to_energies(
                    ham, wfs, self.constraints, use_eps=True)

            self.set_ref_orbitals_and_a_vec(wfs)

    def set_ref_orbitals_and_a_vec(self, wfs):
        self.dm_helper.set_reference_orbitals(wfs, self.n_dim)
        for kpt in wfs.kpt_u:
            u = self.kpointval(kpt)
            self.a_vec_u[u] = np.zeros_like(self.a_vec_u[u])

    def reset(self):
        self.dm_helper = None
        self.error = np.inf
        self.initialized = False
        self.searchdir_algo.reset()

    def todict(self):
        ret = {'name': self.name,
               'searchdir_algo': self.searchdir_algo.todict(),
               'linesearch_algo': self.line_search.todict(),
               'update_ref_orbs_counter': self.update_ref_orbs_counter,
               'update_ref_orbs_canonical': self.update_ref_orbs_canonical,
               'update_precond_counter': self.update_precond_counter,
               'use_prec': self.use_prec,
               'matrix_exp': self.matrix_exp,
               'representation': self.representation,
               'functional': self.func_settings,
               'orthonormalization': self.orthonormalization,
               'randomizeorbitals': self.randomizeorbitals,
               'checkgraderror': self.checkgraderror,
               'localizationtype': self.localizationtype,
               'localizationseed': self.localizationseed,
               'need_localization': self.need_localization,
               'need_init_orbs': self.need_init_orbs,
               'constraints': self.constraints,
               'subspace_convergence': self.subspace_convergence,
               'excited_state': self.excited_state
               }
        if self.gmf:
            ret['partial_diagonalizer'] = \
                self.searchdir_algo.partial_diagonalizer.todict()
        return ret

    def rotate_wavefunctions(self, wfs, a_vec_u, c_ref):

        """
        Apply unitary transformation U = exp(A) to
        the orbitals c_ref

        :param wfs:
        :param a_vec_u:
        :param c_ref:
        :return:
        """

        with wfs.timer('Unitary rotation'):
            for kpt in wfs.kpt_u:
                k = self.kpointval(kpt)
                if self.n_dim[k] == 0:
                    continue

                u_nn = self.get_exponential_matrix_kpt(wfs, kpt, a_vec_u)

                self.dm_helper.appy_transformation_kpt(
                    wfs, u_nn.T, kpt, c_ref[k], False, False)

                with wfs.timer('Calculate projections'):
                    self.dm_helper.update_projections(wfs, kpt)

    def get_exponential_matrix_kpt(self, wfs, kpt, a_vec_u):
        """
        Get unitary matrix U as the exponential of a skew-Hermitian
        matrix A (U = exp(A))
        """

        k = self.kpointval(kpt)

        if self.gd.comm.rank == 0:
            if self.matrix_exp == 'egdecomp-u-invar' and \
                    self.representation == 'u-invar':
                n_occ = get_n_occ(kpt)[0]
                n_v = self.n_dim[k] - n_occ
                a_mat = a_vec_u[k].reshape(n_occ, n_v)
            else:
                a_mat = vec2skewmat(a_vec_u[k], self.n_dim[k],
                                    self.ind_up[k], self.dtype)

            if self.matrix_exp == 'pade-approx':
                # this function takes a lot of memory
                # for large matrices... what can we do?
                with wfs.timer('Pade Approximants'):
                    u_nn = np.ascontiguousarray(expm(a_mat))
            elif self.matrix_exp == 'egdecomp':
                # this method is based on diagonalization
                with wfs.timer('Eigendecomposition'):
                    u_nn, evecs, evals = expm_ed(a_mat, evalevec=True)
            elif self.matrix_exp == 'egdecomp-u-invar':
                with wfs.timer('Eigendecomposition'):
                    u_nn = expm_ed_unit_inv(a_mat, oo_vo_blockonly=False)

        with wfs.timer('Broadcast u_nn'):
            if self.gd.comm.rank != 0:
                u_nn = np.zeros(shape=(self.n_dim[k], self.n_dim[k]),
                                dtype=wfs.dtype)
            self.gd.comm.broadcast(u_nn, 0)

        if self.matrix_exp == 'egdecomp':
            with wfs.timer('Broadcast evecs and evals'):
                if self.gd.comm.rank != 0:
                    evecs = np.zeros(shape=(self.n_dim[k], self.n_dim[k]),
                                     dtype=complex)
                    evals = np.zeros(shape=self.n_dim[k],
                                     dtype=float)
                self.gd.comm.broadcast(evecs, 0)
                self.gd.comm.broadcast(evals, 0)
                self.evecs[k], self.evals[k] = evecs, evals

        return u_nn

    def check_assertions(self, wfs, dens):

        assert dens.mixer.driver.basemixerclass.name == 'no-mixing', \
            'Please, use: mixer={\'backend\': \'no-mixing\'}'
        if wfs.occupations.name != 'mom':
            errormsg = \
                'Please, use occupations={\'name\': \'fixed-uniform\'}'
            assert wfs.occupations.name == 'fixed-uniform', errormsg

    def initialize_mom_reference_orbitals(self, wfs, dens):
        # Reinitialize the MOM reference orbitals
        # after orthogonalization/localization
        occ_name = getattr(wfs.occupations, 'name', None)
        if occ_name == 'mom':
            wfs.occupations.initialize_reference_orbitals()
            wfs.calculate_occupation_numbers(dens.fixed)

    def initial_sort_orbitals_mom(self, wfs):
        occ_name = getattr(wfs.occupations, "name", None)
        if occ_name == 'mom':
            self.initial_occupation_numbers = wfs.occupations.numbers.copy()
            sort_orbitals_according_to_occ(
                wfs, self.constraints, update_mom=True)

    def check_mom(self, wfs, dens):
        occ_name = getattr(wfs.occupations, "name", None)
        if occ_name == 'mom':
            wfs.calculate_occupation_numbers(dens.fixed)
            self.restart = sort_orbitals_according_to_occ(
                wfs, self.constraints, update_mom=True)
            return self.restart
        else:
            return False

    def randomize_orbitals_kpt(self, wfs, kpt):
        """
        Add random noise to orbitals but keep them orthonormal
        """
        nst = self.nbands
        wt = kpt.weight * 0.01
        arand = wt * random_a((nst, nst),
                              wfs.dtype,
                              rng=self.randomizeorbitals)
        arand = arand - arand.T.conj()
        wfs.gd.comm.broadcast(arand, 0)
        self.dm_helper.appy_transformation_kpt(wfs, expm(arand), kpt)

    def calculate_hamiltonian_matrix(self, hamiltonian, wfs, kpt):
        H_MM = self.dm_helper.calculate_hamiltonian_matrix(
            hamiltonian, wfs, kpt)
        return H_MM

    def kpointval(self, kpt):
        return self.nkpts * kpt.s + kpt.q

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, e):
        self._error = e


def vec2skewmat(a_vec, dim, ind_up, dtype):

    a_mat = np.zeros(shape=(dim, dim), dtype=dtype)
    a_mat[ind_up] = a_vec
    a_mat -= a_mat.T.conj()
    np.fill_diagonal(a_mat, a_mat.diagonal() * 0.5)
    return a_mat


def convert_constraints(constraints, n_dim, n_occ, representation):
    """
    Parses and checks the user input of constraints. If constraints are passed
    as a list of a single orbital index all pairs of orbitals involving this
    index are added to the constraints.

    :param constraints: List of constraints for one K-point
    :param n_dim:
    :param n_occ:
    :param representation: Unitary invariant, sparse or full representation
                           determining the electronic degrees of freedom that
                           need to be constrained

    :return: Converted list of constraints
    """

    new = constraints.copy()
    for con in constraints:
        assert isinstance(con, list) or isinstance(con, int), \
            'Check constraints.'
        if isinstance(con, list):
            assert len(con) < 3, 'Check constraints.'
            if len(con) == 1:
                con = con[0]  # List of single index
            else:
                # Make first index always smaller than second index
                if representation != 'full' and con[1] < con[0]:
                    temp = deepcopy(con[0])
                    con[0] = deepcopy(con[1])
                    con[1] = temp
                check_indices(
                    con[0], con[1], n_dim, n_occ, representation)
                continue
        # Add all pairs containing the single index
        if isinstance(con, int):
            new += find_all_pairs(con, n_dim, n_occ, representation)

    # Delete all list containing a single index
    done = False
    while not done:
        done = True
        for i in range(len(new)):
            if len(new[i]) < 2:
                del new[i]
                done = False
                break
    return new


def check_indices(ind1, ind2, n_dim, n_occ, representation):
    """
    Makes sure the user input for the constraints makes sense.
    """

    assert ind1 != ind2, 'Check constraints.'
    if representation == 'full':
        assert ind1 < n_dim and ind2 < n_dim, 'Check constraints.'
    elif representation == 'sparse':
        assert ind1 < n_occ and ind2 < n_dim, 'Check constraints.'
    elif representation == 'u-invar':
        assert ind1 < n_occ and ind2 >= n_occ and ind2 < n_dim, \
            'Check constraints.'


def find_all_pairs(ind, n_dim, n_occ, representation):
    """
    Creates a list of all orbital pairs corresponding to degrees of freedom of
    the system containing an orbital index.

    :param ind: The orbital index
    :param n_dim:
    :param n_occ:
    :param representation: Unitary invariant, sparse or full, defining what
                           index pairs correspond to degrees of freedom of the
                           system
    """

    pairs = []
    if representation == 'u-invar':
        # Only ov rotations are degrees of freedom
        if ind < n_occ:
            for i in range(n_occ, n_dim):
                pairs.append([ind, i])
        else:
            for i in range(n_occ):
                pairs.append([i, ind])
    else:
        if (ind < n_occ and representation == 'sparse') \
                or representation == 'full':
            # oo and ov rotations are degrees of freedom
            for i in range(n_dim):
                if i == ind:
                    continue
                pairs.append([i, ind] if i < ind else [ind, i])
                if representation == 'full':
                    # The orbital rotation matrix is not assumed to be
                    # antihermitian, so the reverse order of indices must be
                    # added as a second constraint
                    pairs.append([ind, i] if i < ind else [i, ind])
        else:
            for i in range(n_occ):
                pairs.append([i, ind])
    return pairs



from __future__ import annotations
from typing import TYPE_CHECKING

from ase.parallel import parprint
from ase.utils.timing import Timer
from pathlib import Path
import numpy as np

from gpaw.new.ase_interface import ASECalculator
from gpaw.nlopt.basic import NLOData
from gpaw.utilities.progressbar import ProgressBar

if TYPE_CHECKING:
    from gpaw.nlopt.adapters import CollinearGSInfo, NoncollinearGSInfo
    from gpaw.typing import ArrayND


def get_mml(gs: CollinearGSInfo | NoncollinearGSInfo,
            spin: int,
            ni: int,
            nf: int,
            timer: Timer | None = None) -> ArrayND:
    """
    Compute momentum matrix elements.

    Parameters
    ----------
    gs
        Ground state adapter.
    spin
        Spin channel index (for spin-polarized systems 0 or 1).
    ni, nf
        First and last band to compute the mml (0 to nb).
    timer
        Timer to keep track of time.

    Returns
    -------
    p_kvnn
        Momentum matrix elements in atomic units gathered on master.
    """

    # Start the timer
    if timer is None:
        timer = Timer()
    parprint(f'Calculating momentum matrix elements for spin channel {spin}.')

    # Specify desired range and number of bands in calculation
    bands = slice(ni, nf)
    nb = nf - ni

    # Spin input
    assert spin < gs.ns, 'Wrong spin input'

    # Parallelisation and memory estimate
    ibzwfs = gs.ibzwfs
    kpt_comm = ibzwfs.kpt_comm
    rank = kpt_comm.rank
    master = (rank == 0)

    nk = len(ibzwfs.rank_k)  # Total number of k-points
    k_q = np.array(list(ibzwfs.q_k.keys()), int)  # k-index for each q-index
    nq = len(k_q)  # Number of k-points (q-indices) for each core
    est_mem = 2 * 3 * nk * nb**2 * 16 / 2**20
    parprint(f'At least {est_mem:.2f} MB of memory is required on master.')

    # Allocate the matrix elements
    p_qvnn = np.empty((nq, 3, nb, nb), dtype=complex)

    # Initial call to print 0 % progress
    if master:
        pb = ProgressBar()

    # Calculate matrix elements in loop over k-points
    for wfs_s in ibzwfs.wfs_qs:
        wfs = gs.get_wfs(wfs_s, spin)

        with timer('Contribution from pseudo wave functions'):
            G_plus_k_Gv, u_nG = gs.get_plane_wave_coefficients(
                wfs, bands=bands, spin=spin)
            p_vnn = np.einsum('Gv,mG,nG->vmn',
                              G_plus_k_Gv, u_nG.conj(), u_nG) * gs.ucvol

        with timer('Contribution from PAW corrections'):
            P_ani = gs.get_wave_function_projections(
                wfs, bands=bands, spin=spin)
            for P_ni, nabla_iiv in zip(P_ani.values(), gs.nabla_aiiv):
                p_vnn -= 1j * np.einsum('mi,nj,ijv->vmn',
                                        P_ni.conj(), P_ni, nabla_iiv)

        p_qvnn[wfs.q] = p_vnn

        if master:
            pb.update(wfs.q / nq)

    if master:
        pb.finish()

    with timer('Gather the data to master'):
        if not master:
            kpt_comm.send(np.array(nq, int), 0)
            kpt_comm.send(k_q, 0)
            kpt_comm.send(p_qvnn, 0)
        else:
            p_kvnn = np.empty((nk, 3, nb, nb), complex)
            p_kvnn[k_q] = p_qvnn
            for gather_rank in range(1, kpt_comm.size):
                _nq = np.empty(1, int)  # We can only communicate numpy arrays
                kpt_comm.receive(_nq, gather_rank)
                nq = _nq[0]

                k_q = np.empty(nq, int)
                kpt_comm.receive(k_q, gather_rank)

                p_qvnn = np.empty((nq, 3, nb, nb), complex)
                kpt_comm.receive(p_qvnn, gather_rank)
                p_kvnn[k_q] = p_qvnn

    # Print the timing
    if master:
        timer.write()

    if rank == 0:
        return p_kvnn
    else:
        return np.array([], dtype=complex)


def make_nlodata(calc: ASECalculator | str | Path,
                 spin_string: str = 'all',
                 ni: int | None = None,
                 nf: int | None = None) -> NLOData:
    """
    This function calculates and returns all required
    NLO data: w_sk, f_skn, E_skn, p_skvnn.

    Parameters
    ----------
    calc
        Calculator or string/path pointing to a .gpw file.
    spin_string
        String denoting which spin channels to include ('all', 's0' , 's1').
    ni
        First band to compute the mml.
    nf
        Last band to compute the mml (relative to number of bands for nf <= 0).

    Returns
    -------
    NLOData
        Data object carrying required matrix elements for NLO calculations.

    """

    if not isinstance(calc, ASECalculator):
        if not (isinstance(calc, str) or isinstance(calc, Path)):
            raise TypeError('Input must be a calculator or a string / path'
                            'pointing to a calculator.')
        from gpaw.new.ase_interface import GPAW
        calc = GPAW(calc, txt=None, parallel={'domain': 1, 'band': 1})
    assert not calc.symmetry.point_group, \
        'Point group symmetry should be off.'

    gs: CollinearGSInfo | NoncollinearGSInfo
    if calc.dft.state.density.collinear:
        from gpaw.nlopt.adapters import CollinearGSInfo
        gs = CollinearGSInfo(calc)
    else:
        from gpaw.nlopt.adapters import NoncollinearGSInfo
        gs = NoncollinearGSInfo(calc)

    # Parse spin string
    ns = gs.ns
    if spin_string == 'all':
        spins = list(range(ns))
    elif spin_string == 's0':
        spins = [0]
    elif spin_string == 's1':
        spins = [1]
        assert spins[0] < ns, 'Wrong spin input'
    else:
        raise NotImplementedError

    # Parse band input
    ibzwfs = gs.ibzwfs
    nb_full = ibzwfs.nbands
    ni = int(ni) if ni is not None else 0
    nf = int(nf) if nf is not None else nb_full
    nf = nb_full + nf if (nf <= 0) else nf

    # Start the timer
    timer = Timer()

    # Get the energy and Fermi-Dirac occupations (data is only in master)
    with timer('Get energies and fermi levels'):

        E_skn, f_skn = ibzwfs.get_all_eigs_and_occs()

        w_sk = np.array([ibzwfs.ibz.weight_k for _ in range(gs.ndensities)])
        w_sk *= gs.bzvol * ibzwfs.spin_degeneracy

    # Compute the momentum matrix elements
    with timer('Compute the momentum matrix elements'):
        p_skvnn = []
        for spin in spins:
            p_kvnn = get_mml(gs=gs, ni=ni, nf=nf,
                             spin=spin, timer=timer)
            p_skvnn.append(p_kvnn)
        if not gs.collinear:
            p_skvnn = [p_skvnn[0] + p_skvnn[1]]

    # Save the output to the file
    return NLOData(w_sk=w_sk,
                   f_skn=f_skn[:, :, ni:nf],
                   E_skn=E_skn[:, :, ni:nf],
                   p_skvnn=np.array(p_skvnn, complex),
                   comm=ibzwfs.kpt_comm)


def get_rml(E_n, p_vnn, pol_v, Etol=1e-6):
    """
    Compute the position matrix elements

    Parameters
    ----------
    E_n
        Band energies.
    p_vnn
        Momentum matrix elements.
    pol_v
        Tensor element.
    Etol
        Tolerance in energy to consider degeneracy.

    Returns
    -------
    r_vnn
        Position matrix elements.
    D_vnn
        Velocity difference matrix elements.

    """

    # Useful variables
    nb = len(E_n)
    r_vnn = np.zeros((3, nb, nb), complex)
    D_vnn = np.zeros((3, nb, nb), complex)
    E_nn = np.tile(E_n[:, None], (1, nb)) - \
        np.tile(E_n[None, :], (nb, 1))
    zeroind = np.abs(E_nn) < Etol
    E_nn[zeroind] = 1
    # Loop over components
    for v1 in set(pol_v):
        r_vnn[v1] = p_vnn[v1] / (1j * E_nn)
        r_vnn[v1, zeroind] = 0
        p_n = np.diag(p_vnn[v1])
        D_vnn[v1] = np.tile(p_n[:, None], (1, nb)) - \
            np.tile(p_n[None, :], (nb, 1))

    return r_vnn, D_vnn


def get_derivative(E_n, r_vnn, D_vnn, pol_v, Etol=1e-6):
    """
    Compute the generalised derivative of position matrix elements

    Parameters
    ----------
    E_n
        Band energies.
    r_vnn
        Momentum matrix elements.
    D_vnn
        Velocity difference matrix elements.
    pol_v
        Tensor element.
    Etol
        Tolerance in energy to consider degeneracy.

    Returns
    -------
    rd_vvnn
        Generalised derivative of position matrix elements.

    """

    # Useful variables
    nb = len(E_n)
    rd_vvnn = np.zeros((3, 3, nb, nb), complex)
    E_nn = np.tile(E_n[:, None], (1, nb)) - \
        np.tile(E_n[None, :], (nb, 1))
    zeroind = np.abs(E_nn) < Etol
    E_nn[zeroind] = 1
    for v1 in set(pol_v):
        for v2 in set(pol_v):
            tmp = (r_vnn[v1] * np.transpose(D_vnn[v2])
                   + r_vnn[v2] * np.transpose(D_vnn[v1])
                   + 1j * np.dot(r_vnn[v1], r_vnn[v2] * E_nn)
                   - 1j * np.dot(r_vnn[v2] * E_nn, r_vnn[v1])) / E_nn
            tmp[zeroind] = 0
            rd_vvnn[v1, v2] = tmp

    return rd_vvnn



from ase import Atom, Atoms

from gpaw import GPAW, Mixer, RMMDIIS
import pytest


def test_eigen_blocked_rmm_diis(in_tmp_dir):
    a = 4.0
    n = 20
    d = 1.0
    x = d / 3**0.5
    atoms = Atoms([Atom('C', (0.0, 0.0, 0.0)),
                   Atom('H', (x, x, x)),
                   Atom('H', (-x, -x, x)),
                   Atom('H', (x, -x, -x)),
                   Atom('H', (-x, x, -x))],
                  cell=(a, a, a), pbc=True)
    base_params = dict(
        mode='fd',
        gpts=(n, n, n),
        nbands=4,
        mixer=Mixer(0.25, 3, 1))
    calc = GPAW(**base_params, txt='a.txt', eigensolver='rmm-diis')
    atoms.calc = calc
    e0 = atoms.get_potential_energy()
    niter0 = calc.get_number_of_iterations()

    es = RMMDIIS(blocksize=3)
    calc = GPAW(**base_params, txt='b.txt', eigensolver=es)
    atoms.calc = calc
    e1 = atoms.get_potential_energy()
    niter1 = calc.get_number_of_iterations()
    assert e0 == pytest.approx(e1, abs=0.000001)
    assert niter0 == pytest.approx(niter1, abs=0)



import numpy as np
import pytest
from gpaw import GPAW
from gpaw.inducedfield.inducedfield_base import BaseInducedField
from gpaw.inducedfield.inducedfield_fdtd import (
    FDTDInducedField, calculate_hybrid_induced_field)
from gpaw.inducedfield.inducedfield_tddft import TDDFTInducedField
from gpaw.tddft import TDDFT, DipoleMomentWriter


@pytest.mark.later
def test_fdtd_ed_inducedfield(in_tmp_dir, gpw_files):
    do_print_values = 0  # Use this for printing the reference values

    # Accuracy
    energy_eps = 0.0005
    poisson_eps = 1e-12
    density_eps = 1e-6

    # load gpw file
    gs_calc = GPAW(gpw_files['na2_isolated'])
    energy = gs_calc.get_potential_energy()

    # Test ground state
    assert energy == pytest.approx(
        -0.631881,
        abs=energy_eps * gs_calc.get_number_of_electrons())

    # Test floating point arithmetic errors
    assert gs_calc.hamiltonian.poisson.shift_indices_1 == pytest.approx(
        [4, 4, 10], abs=0)
    assert gs_calc.hamiltonian.poisson.shift_indices_2 == pytest.approx(
        [8, 8, 16], abs=0)

    # Initialize TDDFT and FDTD
    kick = [0.0, 0.0, 1.0e-3]
    time_step = 10.0
    iterations = 10

    td_calc = TDDFT(gpw_files['na2_isolated'])
    DipoleMomentWriter(td_calc, 'dm.dat')

    # Attach InducedFields to the calculation
    frequencies = [2.05, 4.0]
    width = 0.15
    cl_ind = FDTDInducedField(paw=td_calc,
                              frequencies=frequencies,
                              width=width)
    qm_ind = TDDFTInducedField(paw=td_calc,
                               frequencies=frequencies,
                               width=width)

    # Propagate TDDFT and FDTD
    td_calc.absorption_kick(kick_strength=kick)
    td_calc.propagate(time_step, iterations // 2)
    td_calc.write('td.gpw', 'all')
    cl_ind.write('cl.ind')
    qm_ind.write('qm.ind')

    # Restart
    td_calc = TDDFT('td.gpw')
    DipoleMomentWriter(td_calc, 'dm.dat')
    cl_ind = FDTDInducedField(filename='cl.ind',
                              paw=td_calc)
    qm_ind = TDDFTInducedField(filename='qm.ind',
                               paw=td_calc)
    td_calc.propagate(time_step, iterations // 2)
    td_calc.write('td.gpw', 'all')
    cl_ind.write('cl.ind')
    qm_ind.write('qm.ind')

    # Test
    ref_cl_dipole_moment = [5.25374117e-14, 5.75811267e-14, 3.08349334e-02]
    ref_qm_dipole_moment = [1.78620337e-11, -1.57782578e-11, 5.21368300e-01]

    tol = 1e-4
    assert td_calc.hamiltonian.poisson.get_classical_dipole_moment() == (
        pytest.approx(ref_cl_dipole_moment, abs=tol))
    assert td_calc.hamiltonian.poisson.get_quantum_dipole_moment() == (
        pytest.approx(ref_qm_dipole_moment, abs=tol))

    # Calculate induced fields
    td_calc = TDDFT('td.gpw')

    # Classical subsystem
    cl_ind = FDTDInducedField(filename='cl.ind', paw=td_calc)
    cl_ind.calculate_induced_field(gridrefinement=2)
    cl_ind.write('cl_field.ind', mode='all')

    # Quantum subsystem
    qm_ind = TDDFTInducedField(filename='qm.ind', paw=td_calc)
    qm_ind.calculate_induced_field(gridrefinement=2)
    qm_ind.write('qm_field.ind', mode='all')

    # Total system, interpolate/extrapolate to a grid with spacing h
    tot_ind = calculate_hybrid_induced_field(cl_ind, qm_ind, h=0.4)
    tot_ind.write('tot_field.ind', mode='all')

    # Test induced fields
    ref_values = [72404.467117024149,
                  0.520770766296,
                  0.520770766299,
                  0.830247064075,
                  72416.234345610734,
                  0.517294132489,
                  0.517294132492,
                  0.824704513888,
                  2451.767847927681,
                  0.088037476748,
                  0.088037476316,
                  0.123334033914,
                  2454.462292798476,
                  0.087537484422,
                  0.087537483971,
                  0.122592730690,
                  76582.089818637178,
                  0.589941751987,
                  0.589941751804,
                  0.869526245360,
                  76592.175846021099,
                  0.586223386358,
                  0.586223386102,
                  0.864478308364]

    for fname in ['cl_field.ind', 'qm_field.ind', 'tot_field.ind']:
        ind = BaseInducedField(filename=fname,
                               readmode='field')
        # Estimate tolerance (worst case error accumulation)
        tol = (iterations * ind.fieldgd.integrate(ind.fieldgd.zeros() + 1.0) *
               max(density_eps, np.sqrt(poisson_eps)))
        if do_print_values:
            print('tol = %.12f' % tol)
        for w in range(len(frequencies)):
            val = ind.fieldgd.integrate(ind.Ffe_wg[w])
            assert val == pytest.approx(ref_values.pop(0), abs=tol)
            for v in range(3):
                val = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[w][v]))
                assert val == pytest.approx(ref_values.pop(0), abs=tol)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""
Python wrapper functions for the ``C`` package:
Basic Linear Algebra Subroutines (BLAS)

See also:
http://en.wikipedia.org/wiki/Basic_Linear_Algebra_Subprograms
and
http://www.netlib.org/lapack/lug/node145.html
"""
from typing import TypeVar

import gpaw.cgpaw as cgpaw
import numpy as np
import scipy.linalg.blas as blas
from gpaw import debug
from gpaw.new import prod
from gpaw.typing import Array2D, ArrayND
from gpaw.utilities import is_contiguous


def is_finite(array, tril=False):
    if isinstance(array, np.ndarray):
        xp = np
    else:
        from gpaw.gpu import cupy as xp
    if tril:
        array = xp.tril(array)
    return xp.isfinite(array).all()


__all__ = ['mmm']

T = TypeVar('T', float, complex)


def mmm(alpha: T,
        a: Array2D,
        opa: str,
        b: Array2D,
        opb: str,
        beta: T,
        c: Array2D) -> None:
    """Matrix-matrix multiplication using dgemm or zgemm.

    For opa='N' and opb='N', we have:::

        c <- αab + βc.

    Use 'T' to transpose matrices and 'C' to transpose and complex conjugate
    matrices.
    """

    assert opa in 'NTC'
    assert opb in 'NTC'

    if opa == 'N':
        a1, a2 = a.shape
    else:
        a2, a1 = a.shape
    if opb == 'N':
        b1, b2 = b.shape
    else:
        b2, b1 = b.shape
    assert a2 == b1
    assert c.shape == (a1, b2)

    assert a.dtype == b.dtype == c.dtype
    assert a.strides[1] == c.itemsize or a.size == 0
    assert b.strides[1] == c.itemsize or b.size == 0
    assert c.strides[1] == c.itemsize or c.size == 0
    if a.dtype == float:
        assert not isinstance(alpha, complex)
        assert not isinstance(beta, complex)
    else:
        assert a.dtype == complex

    cgpaw.mmm(alpha, a, opa, b, opb, beta, c)


def gpu_mmm(alpha, a, opa, b, opb, beta, c):
    """Launch CPU or GPU version of mmm()."""
    m = b.shape[1] if opb == 'N' else b.shape[0]
    n = a.shape[0] if opa == 'N' else a.shape[1]
    k = b.shape[0] if opb == 'N' else b.shape[1]
    lda = a.strides[0] // a.itemsize
    ldb = b.strides[0] // b.itemsize
    ldc = c.strides[0] // c.itemsize
    cgpaw.mmm_gpu(alpha, a.data.ptr, lda, opa,
                  b.data.ptr, ldb, opb, beta,
                  c.data.ptr, ldc, c.itemsize,
                  m, n, k)


def gpu_scal(alpha, x):
    """alpha x

    Performs the operation::

      x <- alpha * x

    """
    if debug:
        if isinstance(alpha, complex):
            assert is_contiguous(x, complex)
        else:
            assert isinstance(alpha, float)
            assert x.dtype in [float, complex]
            assert x.flags.c_contiguous
    cgpaw.scal_gpu(alpha, x.data.ptr, x.shape, x.dtype)


def to2d(array: ArrayND) -> Array2D:
    """2D view af ndarray.

    >>> to2d(np.zeros((2, 3, 4))).shape
    (2, 12)
    """
    shape = array.shape
    return array.reshape((shape[0], prod(shape[1:])))


def mmmx(alpha: T,
         a: ArrayND,
         opa: str,
         b: ArrayND,
         opb: str,
         beta: T,
         c: ArrayND) -> None:
    """Matrix-matrix multiplication using dgemm or zgemm.

    Arrays a, b and c are converted to 2D arrays before calling mmm().
    """
    mmm(alpha, to2d(a), opa, to2d(b), opb, beta, to2d(c))


def gpu_gemm(alpha, a, b, beta, c, transa='n'):
    """General Matrix Multiply.

    Performs the operation::

      c <- alpha * b.a + beta * c

    If transa is "n", ``b.a`` denotes the matrix multiplication defined by::

                      _
                     \
      (b.a)        =  ) b  * a
           ijkl...   /_  ip   pjkl...
                      p

    If transa is "t" or "c", ``b.a`` denotes the matrix multiplication
    defined by::

                      _
                     \
      (b.a)        =  ) b    *    a
           ij        /_  iklm...   jklm...
                     klm...

    where in case of "c" also complex conjugate of a is taken.
    """
    if debug:
        assert beta == 0.0 or is_finite(c)

        assert (a.dtype == float and b.dtype == float and c.dtype == float and
                isinstance(alpha, float) and isinstance(beta, float) or
                a.dtype == complex and b.dtype == complex and
                c.dtype == complex)
        assert a.flags.c_contiguous
        if transa == 'n':
            assert c.flags.c_contiguous or (c.ndim == 2
                                            and c.strides[1] == c.itemsize)
            assert b.ndim == 2
            assert b.strides[1] == b.itemsize
            assert a.shape[0] == b.shape[1]
            assert c.shape == b.shape[0:1] + a.shape[1:]
        else:
            assert b.size == 0 or b[0].flags.c_contiguous
            assert c.strides[1] == c.itemsize
            assert a.shape[1:] == b.shape[1:]
            assert c.shape == (b.shape[0], a.shape[0])

    cgpaw.gemm_gpu(alpha, a.data.ptr, a.shape,
                   b.data.ptr, b.shape, beta,
                   c.data.ptr, c.shape,
                   a.dtype, transa)


def gpu_gemv(alpha, a, x, beta, y, trans='t'):
    """General Matrix Vector product.

    Performs the operation::

      y <- alpha * a.x + beta * y

    ``a.x`` denotes matrix multiplication, where the product-sum is
    over the entire length of the vector x and
    the first dimension of a (for trans='n'), or
    the last dimension of a (for trans='t' or 'c').

    If trans='c', the complex conjugate of a is used. The default is
    trans='t', i.e. behaviour like np.dot with a 2D matrix and a vector.
    """
    if debug:
        assert (a.dtype == float and x.dtype == float and y.dtype == float and
                isinstance(alpha, float) and isinstance(beta, float) or
                a.dtype == complex and x.dtype == complex and
                y.dtype == complex)
        assert a.flags.c_contiguous
        assert y.flags.c_contiguous
        assert x.ndim == 1
        assert y.ndim == a.ndim - 1
        if trans == 'n':
            assert a.shape[0] == x.shape[0]
            assert a.shape[1:] == y.shape
        else:
            assert a.shape[-1] == x.shape[0]
            assert a.shape[:-1] == y.shape

    cgpaw.gemv_gpu(alpha, a.data.ptr, a.shape,
                   x.data.ptr, x.shape, beta,
                   y.data.ptr, a.dtype,
                   trans)


def axpy(alpha, x, y):
    """alpha x plus y.

    Performs the operation::

      y <- alpha * x + y

    """
    if x.size == 0:
        return
    assert x.flags.contiguous
    assert y.flags.contiguous
    x = x.ravel()
    y = y.ravel()
    if x.dtype == float:
        z = blas.daxpy(x, y, a=alpha)
    else:
        z = blas.zaxpy(x, y, a=alpha)
    assert z is y, (x, y, x.shape, y.shape)


def gpu_axpy(alpha, x, y):
    """alpha x plus y.

    Performs the operation::

      y <- alpha * x + y

    """
    if debug:
        if isinstance(alpha, complex):
            assert is_contiguous(x, complex) and is_contiguous(y, complex)
        else:
            assert isinstance(alpha, float)
            assert x.dtype in [float, complex]
            assert x.dtype == y.dtype
            assert x.flags.c_contiguous and y.flags.c_contiguous
        assert x.shape == y.shape

    cgpaw.axpy_gpu(alpha, x.data.ptr, x.shape,
                   y.data.ptr, y.shape,
                   x.dtype)


def rk(alpha, a, beta, c, trans='c'):
    """Rank-k update of a matrix.

    For ``trans='c'`` the following operation is performed:::

              †
      c <- αaa + βc,

    and for ``trans='t'`` we get:::

             †
      c <- αa a + βc

    If the ``a`` array has more than 2 dimensions then the 2., 3., ...
    axes are combined.

    Only the lower triangle of ``c`` will contain sensible numbers.
    """
    if debug:
        assert beta == 0.0 or is_finite(c, tril=True)

        assert (a.dtype == float and c.dtype == float or
                a.dtype == complex and c.dtype == complex)
        assert a.flags.c_contiguous
        assert a.ndim > 1
        if trans == 'n':
            assert c.shape == (a.shape[1], a.shape[1])
        else:
            assert c.shape == (a.shape[0], a.shape[0])
        assert c.strides[1] == c.itemsize or c.size == 0

    cgpaw.rk(alpha, a, beta, c, trans)


def gpu_rk(alpha, a, beta, c, trans='c'):
    """Launch CPU or GPU version of rk()."""
    cgpaw.rk_gpu(alpha, a.data.ptr, a.shape,
                 beta, c.data.ptr, c.shape,
                 a.dtype)


def r2k(alpha, a, b, beta, c, trans='c'):
    """Rank-2k update of a matrix.

    Performs the operation::

                        dag        cc       dag
      c <- alpha * a . b    + alpha  * b . a    + beta * c

    or if trans='n'::
                    dag           cc   dag
      c <- alpha * a   . b + alpha  * b   . a + beta * c

    where ``a.b`` denotes the matrix multiplication defined by::

                 _
                \
      (a.b)   =  ) a         * b
           ij   /_  ipklm...     pjklm...
               pklm...

    ``cc`` denotes complex conjugation.

    ``dag`` denotes the hermitian conjugate (complex conjugation plus a
    swap of axis 0 and 1).

    Only the lower triangle of ``c`` will contain sensible numbers.
    """
    if debug:
        assert beta == 0.0 or is_finite(c, tril=True)
        assert (a.dtype == float and b.dtype == float and c.dtype == float or
                a.dtype == complex and b.dtype == complex and
                c.dtype == complex)
        assert a.flags.c_contiguous and b.flags.c_contiguous
        assert a.ndim > 1
        assert a.shape == b.shape
        if trans == 'c':
            assert c.shape == (a.shape[0], a.shape[0])
        else:
            assert c.shape == (a.shape[1], a.shape[1])
        assert c.strides[1] == c.itemsize or c.size == 0

    cgpaw.r2k(alpha, a, b, beta, c, trans)


def gpu_r2k(alpha, a, b, beta, c, trans='c'):
    """Launch CPU or GPU version of r2k()."""
    cgpaw.r2k_gpu(alpha, a.data.ptr, a.shape,
                  b.data.ptr, b.shape, beta,
                  c.data.ptr, c.shape,
                  a.dtype)


def gpu_dotc(a, b):
    r"""Dot product, conjugating the first vector with complex arguments.

    Returns the value of the operation::

        _
       \   cc
        ) a       * b
       /_  ijk...    ijk...
       ijk...

    ``cc`` denotes complex conjugation.
    """
    if debug:
        assert ((is_contiguous(a, float) and is_contiguous(b, float)) or
                (is_contiguous(a, complex) and is_contiguous(b, complex)))
        assert a.shape == b.shape

    return cgpaw.dotc_gpu(a.data.ptr, a.shape,
                          b.data.ptr, a.dtype)


def gpu_dotu(a, b):
    """Dot product, NOT conjugating the first vector with complex arguments.

    Returns the value of the operation::

        _
       \
        ) a       * b
       /_  ijk...    ijk...
       ijk...


    """
    if debug:
        assert ((is_contiguous(a, float) and is_contiguous(b, float)) or
                (is_contiguous(a, complex) and is_contiguous(b, complex)))
        assert a.shape == b.shape

    return cgpaw.dotu_gpu(a.data.ptr, a.shape,
                          b.data.ptr, a.dtype)


def _gemmdot(a, b, alpha=1.0, beta=1.0, out=None, trans='n'):
    """Matrix multiplication using gemm.

    return reference to out, where::

      out <- alpha * a . b + beta * out

    If out is None, a suitably sized zero array will be created.

    ``a.b`` denotes matrix multiplication, where the product-sum is
    over the last dimension of a, and either
    the first dimension of b (for trans='n'), or
    the last dimension of b (for trans='t' or 'c').

    If trans='c', the complex conjugate of b is used.
    """
    # Store original shapes
    ashape = a.shape
    bshape = b.shape

    # Vector-vector multiplication is handled by dotu
    if a.ndim == 1 and b.ndim == 1:
        assert out is None
        if trans == 'c':
            return alpha * np.vdot(b, a)  # dotc conjugates *first* argument
        else:
            return alpha * a.dot(b)

    # Map all arrays to 2D arrays
    a = a.reshape(-1, a.shape[-1])
    if trans == 'n':
        b = b.reshape(b.shape[0], -1)
        outshape = a.shape[0], b.shape[1]
    else:  # 't' or 'c'
        b = b.reshape(-1, b.shape[-1])

    # Apply BLAS gemm routine
    outshape = a.shape[0], b.shape[trans == 'n']
    if out is None:
        # (ATLAS can't handle uninitialized output array)
        out = np.zeros(outshape, a.dtype)
    else:
        out = out.reshape(outshape)
    mmmx(alpha, a, 'N', b, trans.upper(), beta, out)

    # Determine actual shape of result array
    if trans == 'n':
        outshape = ashape[:-1] + bshape[1:]
    else:  # 't' or 'c'
        outshape = ashape[:-1] + bshape[:-1]
    return out.reshape(outshape)


if not hasattr(cgpaw, 'mmm'):
    # These are the functions used with noblas=True
    # TODO: move these functions elsewhere so that
    # they can be used for unit tests

    def op(o, m):
        if o.upper() == 'N':
            return m
        if o.upper() == 'T':
            return m.T
        if o.upper() == 'C':
            return m.conj().T
        raise ValueError(f'unknown op: {o}')

    def rk(alpha, a, beta, c, trans='c'):  # noqa
        if c.size == 0:
            return
        if beta == 0:
            c[:] = 0.0
        else:
            c *= beta
        if trans == 'n':
            c += alpha * a.conj().T.dot(a)
        else:
            a = a.reshape((len(a), -1))
            c += alpha * a.dot(a.conj().T)

    def r2k(alpha, a, b, beta, c, trans='c'):  # noqa
        if c.size == 0:
            return
        if beta == 0.0:
            c[:] = 0.0
        else:
            c *= beta
        if trans == 'c':
            c += (alpha * a.reshape((len(a), -1))
                  .dot(b.reshape((len(b), -1)).conj().T) +
                  alpha * b.reshape((len(b), -1))
                  .dot(a.reshape((len(a), -1)).conj().T))
        else:
            c += alpha * (a.conj().T @ b + b.conj().T @ a)

    def mmm(alpha: T, a: np.ndarray, opa: str,  # noqa
            b: np.ndarray, opb: str,
            beta: T, c: np.ndarray) -> None:
        if beta == 0.0:
            c[:] = 0.0
        else:
            c *= beta
        c += alpha * op(opa, a).dot(op(opb, b))

    gemmdot = _gemmdot

elif not debug:
    mmm = cgpaw.mmm  # noqa
    rk = cgpaw.rk  # noqa
    r2k = cgpaw.r2k  # noqa
    gemmdot = _gemmdot

else:
    def gemmdot(a, b, alpha=1.0, beta=1.0, out=None, trans='n'):
        assert a.flags.c_contiguous
        assert b.flags.c_contiguous
        assert a.dtype == b.dtype
        if trans == 'n':
            assert a.shape[-1] == b.shape[0]
        else:
            assert a.shape[-1] == b.shape[-1]
        if out is not None:
            assert out.flags.c_contiguous
            assert a.dtype == out.dtype
            assert a.ndim > 1 or b.ndim > 1
            if trans == 'n':
                assert out.shape == a.shape[:-1] + b.shape[1:]
            else:
                assert out.shape == a.shape[:-1] + b.shape[:-1]
        return _gemmdot(a, b, alpha, beta, out, trans)



import pytest
from gpaw import GPAW
from gpaw.tddft import TDDFT, DipoleMomentWriter


@pytest.mark.later
def test_fdtd_ed(in_tmp_dir, gpw_files):
    # Accuracy
    energy_eps = 0.0005

    # load gpw file
    gs_calc = GPAW(gpw_files['na2_isolated'])
    energy = gs_calc.get_potential_energy()

    # Test ground state
    assert energy == pytest.approx(
        -0.631881, abs=energy_eps * gs_calc.get_number_of_electrons())

    # Test floating point arithmetic errors
    assert gs_calc.hamiltonian.poisson.shift_indices_1 == pytest.approx(
        [4, 4, 10], abs=0)
    assert gs_calc.hamiltonian.poisson.shift_indices_2 == pytest.approx(
        [8, 8, 16], abs=0)

    # Initialize TDDFT and FDTD
    kick = [0.0, 0.0, 1.0e-3]
    time_step = 10.0
    max_time = 100  # 0.1 fs

    td_calc = TDDFT(gpw_files['na2_isolated'])
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick(kick_strength=kick)

    # Propagate TDDFT and FDTD
    td_calc.propagate(time_step, max_time / time_step / 2)
    td_calc.write('td.gpw', 'all')

    td_calc2 = TDDFT('td.gpw')
    DipoleMomentWriter(td_calc2, 'dm.dat')
    td_calc2.propagate(time_step, max_time / time_step / 2)

    # Test
    ref_cl_dipole_moment = [5.25374117e-14, 5.75811267e-14, 3.08349334e-02]
    ref_qm_dipole_moment = [1.78620337e-11, -1.57782578e-11, 5.21368300e-01]

    tol = 1e-4
    assert td_calc2.hamiltonian.poisson.get_classical_dipole_moment() == (
        pytest.approx(ref_cl_dipole_moment, abs=tol))
    assert td_calc2.hamiltonian.poisson.get_quantum_dipole_moment() == (
        pytest.approx(ref_qm_dipole_moment, abs=tol))



import numpy as np
from ase.units import Ha
from gpaw.mpi import world
from gpaw.response import ResponseContext
from gpaw.response.coulomb_kernels import CoulombKernel
from gpaw.response.screened_interaction import initialize_w_calculator
from gpaw.response import timer
from gpaw.response.pw_parallelization import Blocks1D
from gpaw.response.pair import KPointPairFactory
from gpaw.wannier90 import read_uwan


def ibz2bz_map(qd):
    """ Maps each k in BZ to corresponding k in IBZ. """
    out_map = [[] for _ in range(qd.nibzkpts)]
    for iK in range(qd.nbzkpts):
        ik = qd.bz2ibz_k[iK]
        out_map[ik].append(iK)
    return out_map


def initialize_w_model(chi0calc, truncation=None, integrate_gamma=0,
                       q0_correction=False, txt='w_model.out',
                       eta=None, world=world, timer=None):
    """ Helper function to initialize ModelInteraction

    Parameters
    ----------
    chi0calc: Chi0Calculator
    truncation: str
        Coulomb truncation scheme. Can be either 2D, 1D, 0D or None.
    integrate_gamma: int
        Method to integrate the Coulomb interaction. 1 is a numerical
        integration at all q-points with G=[0,0,0] - this breaks the
        symmetry slightly. 0 is analytical integration at q=[0,0,0] only -
        this conserves the symmetry. integrate_gamma=2 is the same as 1,
        but the average is only carried out in the non-periodic directions.
    q0_correction: bool
        Analytic correction to the q=0 contribution applicable to 2D
        systems.
    txt: str
        Filename of output files.
    world: MPI comm
    timer: ResponseContext timer
    """
    gs = chi0calc.gs
    wcontext = ResponseContext(txt=txt,
                               comm=world, timer=timer)
    coulomb = CoulombKernel.from_gs(gs, truncation=truncation)
    if eta is not None:
        eta /= Ha
    wcalc = initialize_w_calculator(chi0calc,
                                    wcontext,
                                    coulomb=coulomb,
                                    xc='RPA',
                                    integrate_gamma=integrate_gamma,
                                    q0_correction=q0_correction,
                                    eta=eta)
    return ModelInteraction(wcalc)


class ModelInteraction:

    def __init__(self, wcalc):
        """Class to compute interaction matrix in Wannier basis

        Parameters
        ----------
        wcalc: WCalculator
        """

        self.wcalc = wcalc
        self.gs = wcalc.gs
        self.context = self.wcalc.context
        self.qd = self.wcalc.qd

    @timer('Calculate W in Wannier')
    def calc_in_Wannier(self, chi0calc, Uwan_mnk, bandrange, spin=0):
        """Calculates the screened interaction matrix in Wannier basis.
        NOTE: Does not work with SOC!

        Parameters:
        ----------
        chi0calc: Chi0Calculator
        Uwan_mnk: cmplx or str
            Wannier transformation matrix. if str name of wannier90 output
            m: Wannier index, n: bandindex, k: k-index
        bandrange: int
            range of bands that the wannier functions were constructed from
        spin: int
            spin index (0 or 1)

        Documentation
        -------------

        W_n1,n2;n3,n4(R=0, w) =
        <w^*_{n1,R=0} w_{n2, R=0} | W(r,r',w) |w^*_{n3,R=0} w_{n4, R=0} >

        w_{n R} = V/(2pi)^3 int_{BZ} dk e^{-kR} psi^w_{nk}
        psi^w_{nk} = sum_m U_nm(k) psi^{KS}_{mk}

        In this implementation we compute W_n1,n2;n3,n4(R=0) efficiently
        by expressing it in terms of the reduced Wannier density matrices

        A_{n1,n2,q} = sum_{k} rhowan^{n1,k}_{n2, k+q}

        where the Wannier density matrix is calculated from the usual
        density matrix
        rho^{m1,k}_{m2 k+q} = < psi_{m1,k} | e^{i(q+G)r} | psi_{m2, k+q} >
        and the Wannier transformation matrices U_{nm}(k) as

        rhowan^{n1, k}_{n2, k+q} =
        = sum_{m1, m5} U^*_{n1,m1}(k) U_{n2, m2}(k+q) rho^{m1,k}_{m2 k+q}
        """

        ibz2bz = ibz2bz_map(self.gs.kd)

        # Create pair_factory with nblocks = 1
        pair_factory = KPointPairFactory(self.gs, self.context)
        pair_calc = pair_factory.pair_calculator()

        if isinstance(Uwan_mnk, str):
            # read w90 transformation matrix from file
            seed = Uwan_mnk
            Uwan_mnk, nk, nwan, nband = read_uwan(seed,
                                                  self.gs.kd,
                                                  dis=False)
            if bandrange[1] - bandrange[0] > nband:
                Uwan_mnk, nk, nwan, nband = read_uwan(seed,
                                                      self.gs.kd,
                                                      dis=True)
        else:
            nk = Uwan_mnk.shape[2]
            nband = Uwan_mnk.shape[1]
            nwan = Uwan_mnk.shape[0]

        assert nk == self.gs.kd.nbzkpts
        assert bandrange[1] - bandrange[0] == nband
        wd = chi0calc.wd
        nfreq = len(wd)

        # Find frequency range for block distributed arrays
        blockdist = chi0calc.chi0_body_calc.get_blockdist()
        blocks1d = Blocks1D(blockdist.blockcomm, nfreq)
        mynfreq = blocks1d.nlocal
        self.intrablockcomm = blockdist.intrablockcomm

        # Variable that will store the screened interaction in Wannier basis
        Wwan_wijkl = np.zeros([mynfreq, nwan, nwan, nwan, nwan], dtype=complex)

        for iq, q_c in enumerate(self.gs.kd.ibzk_kc):
            self.context.print('iq = ', iq, '/', self.gs.kd.nibzkpts)
            # Calculate chi0 and W for IBZ k-point q
            self.context.print('calculating chi0...')
            self.context.timer.start('chi0')
            chi0 = chi0calc.calculate(q_c)
            self.context.timer.stop('chi0')
            qpd = chi0.qpd
            self.context.print('calculating W_wGG...')
            W_wGG = self.wcalc.calculate_W_wGG(chi0,
                                               fxc_mode='GW',
                                               only_correlation=False)
            pawcorr = chi0calc.chi0_body_calc.pawcorr

            self.context.print('Projecting to localized Wannier basis...')
            # Loop over all equivalent k-points
            for iQ in ibz2bz[iq]:
                Q_c = self.wcalc.qd.bzk_kc[iQ]
                assert self.wcalc.qd.where_is_q(Q_c,
                                                self.wcalc.qd.bzk_kc) == iQ

                A_mnG = self.get_reduced_wannier_density_matrix(spin,
                                                                Q_c,
                                                                iq,
                                                                bandrange,
                                                                pawcorr,
                                                                qpd,
                                                                pair_calc,
                                                                pair_factory,
                                                                Uwan_mnk)
                if self.qd.time_reversal_k[iQ]:
                    # TR corresponds to complex conjugation
                    A_mnG = A_mnG.conj()

                Wwan_wijkl += np.einsum('ijk,lkm,pqm->lipjq',
                                        A_mnG.conj(),
                                        W_wGG,
                                        A_mnG,
                                        optimize='optimal')

        # factor from three BZ summations and taking from Hartree to eV
        # and volume factor from matrix element in PW basis
        factor = Ha / self.gs.kd.nbzkpts**3 / self.gs.volume
        Wwan_wijkl *= factor
        self.context.write_timer()

        return wd.omega_w * Ha, blocks1d.all_gather(Wwan_wijkl)

    @timer('get_reduced_wannier_density_matrix')
    def get_reduced_wannier_density_matrix(self, spin, Q_c, iq, bandrange,
                                           pawcorr, qpd, pair_calc,
                                           pair_factory, Uwan_mnk):
        """
        Returns sum_k sum_(m1,m2) U_{n1m1}* U_{n2m2} rho^{m1 k}_{m2 k-q}(G)
        where rho is the usual density matrix and U are wannier tranformation
        matrices.
        """
        nG = qpd.ngmax
        nwan = Uwan_mnk.shape[0]
        A_mnG = np.zeros([nwan, nwan, nG], dtype=complex)

        # Parallell sum over k-points
        for iK1 in self.myKrange():
            kpt1 = pair_factory.get_k_point(spin, iK1,
                                            bandrange[0],
                                            bandrange[1])
            iK2 = self.gs.kd.find_k_plus_q(Q_c, [kpt1.K])[0]
            kpt2 = pair_factory.get_k_point(spin, iK2,
                                            bandrange[0],
                                            bandrange[1])

            # Calculate density matrix
            rholoc, iqloc, sign = self.get_density_matrix(kpt1,
                                                          kpt2,
                                                          pawcorr,
                                                          qpd,
                                                          pair_calc)
            assert iqloc == iq

            # Rotate to Wannier basis and sum to get reduced Wannier
            # density matrix A
            A_mnG += np.einsum('ia,jb,abG->ijG',
                               Uwan_mnk[:, :, iK1].conj(),
                               Uwan_mnk[:, :, iK2],
                               rholoc)
        self.intrablockcomm.sum(A_mnG)
        return A_mnG

    def get_density_matrix(self, kpt1, kpt2, pawcorr, qpd, pair_calc):
        from gpaw.response.g0w0 import QSymmetryOp, get_nmG

        symop, iq = QSymmetryOp.get_symop_from_kpair(
            self.gs.kd, self.qd, kpt1, kpt2)
        nG = qpd.ngmax
        mypawcorr, I_G = symop.apply_symop_q(
            qpd, pawcorr, kpt1, kpt2)

        rho_mnG = np.zeros((len(kpt1.eps_n), len(kpt2.eps_n), nG),
                           complex)
        for m in range(len(rho_mnG)):
            rho_mnG[m] = get_nmG(kpt1, kpt2, mypawcorr, m, qpd, I_G, pair_calc)

        return rho_mnG, iq, symop.sign

    def myKrange(self, rank=None):
        if rank is None:
            rank = self.intrablockcomm.rank
        nK = self.gs.kd.nbzkpts
        myKsize = -(-nK // self.intrablockcomm.size)
        myKrange = range(rank * myKsize,
                         min((rank + 1) * myKsize, nK))
        # myKsize = len(myKrange)
        return myKrange



from ase import Atoms
from gpaw import GPAW, PoissonSolver
import pytest
from ase.units import Bohr, Hartree


def test_xc_revPBE():
    a = 7.5 * Bohr
    n = 16
    atoms = Atoms('He', [(0.0, 0.0, 0.0)], cell=(a, a, a), pbc=True)
    params = dict(mode='fd',
                  gpts=(n, n, n),
                  nbands=1,
                  xc={'name': 'PBE', 'stencil': 1},
                  poissonsolver=PoissonSolver('fd'))
    atoms.calc = GPAW(**params)
    e1 = atoms.get_potential_energy()
    e1a = atoms.calc.get_reference_energy()
    params['xc'] = {'name': 'revPBE', 'stencil': 1}
    atoms.calc = GPAW(**params)
    e2 = atoms.get_potential_energy()
    e2a = atoms.calc.get_reference_energy()

    assert e1a == pytest.approx(-2.893 * Hartree, abs=8e-3)
    assert e2a == pytest.approx(-2.908 * Hartree, abs=9e-3)
    assert e1 == pytest.approx(e2, abs=4e-3)

    energy_tolerance = 0.0005
    assert e1 == pytest.approx(-0.0790449962, abs=energy_tolerance)
    assert e2 == pytest.approx(-0.08147563, abs=energy_tolerance)
